# Copyright (C) 2007-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from .. import bugtracker, urlutils
from . import TestCase, TestCaseWithMemoryTransport


class ErrorsTest(TestCaseWithMemoryTransport):
    def test_unknown_bug_tracker_abbreviation(self):
        """Test the formatting of UnknownBugTrackerAbbreviation."""
        branch = self.make_branch("some_branch")
        error = bugtracker.UnknownBugTrackerAbbreviation("xxx", branch)
        self.assertEqual(
            "Cannot find registered bug tracker called xxx on {}".format(branch),
            str(error),
        )

    def test_malformed_bug_identifier(self):
        """Test the formatting of MalformedBugIdentifier."""
        error = bugtracker.MalformedBugIdentifier("bogus", "reason for bogosity")
        self.assertEqual(
            "Did not understand bug identifier bogus: reason for bogosity. "
            'See "brz help bugs" for more information on this feature.',
            str(error),
        )

    def test_incorrect_url(self):
        err = bugtracker.InvalidBugTrackerURL("foo", "http://bug.example.com/")
        self.assertEqual(
            (
                'The URL for bug tracker "foo" doesn\'t contain {id}: '
                "http://bug.example.com/"
            ),
            str(err),
        )


class TestGetBugURL(TestCaseWithMemoryTransport):
    """Tests for bugtracker.get_bug_url."""

    class TransientTracker:
        """An transient tracker used for testing."""

        @classmethod
        def get(klass, abbreviation, branch):
            klass.log.append(("get", abbreviation, branch))
            if abbreviation != "transient":
                return None
            return klass()

        def get_bug_url(self, bug_id):
            self.log.append(("get_bug_url", bug_id))
            return "http://bugs.example.com/{}".format(bug_id)

    def setUp(self):
        super().setUp()
        self.tracker_type = TestGetBugURL.TransientTracker
        self.tracker_type.log = []
        bugtracker.tracker_registry.register("transient", self.tracker_type)
        self.addCleanup(bugtracker.tracker_registry.remove, "transient")

    def test_get_bug_url_for_transient_tracker(self):
        branch = self.make_branch("some_branch")
        self.assertEqual(
            "http://bugs.example.com/1234",
            bugtracker.get_bug_url("transient", branch, "1234"),
        )
        self.assertEqual(
            [("get", "transient", branch), ("get_bug_url", "1234")],
            self.tracker_type.log,
        )

    def test_unrecognized_abbreviation_raises_error(self):
        """If the abbreviation is unrecognized, then raise an error."""
        branch = self.make_branch("some_branch")
        self.assertRaises(
            bugtracker.UnknownBugTrackerAbbreviation,
            bugtracker.get_bug_url,
            "xxx",
            branch,
            "1234",
        )
        self.assertEqual([("get", "xxx", branch)], self.tracker_type.log)


class TestBuiltinTrackers(TestCaseWithMemoryTransport):
    """Test that the builtin trackers are registered and return sane URLs."""

    def test_launchpad_registered(self):
        """The Launchpad bug tracker should be registered by default and
        generate Launchpad bug page URLs.
        """
        branch = self.make_branch("some_branch")
        tracker = bugtracker.tracker_registry.get_tracker("lp", branch)
        self.assertEqual("https://launchpad.net/bugs/1234", tracker.get_bug_url("1234"))

    def test_debian_registered(self):
        """The Debian bug tracker should be registered by default and generate
        bugs.debian.org bug page URLs.
        """
        branch = self.make_branch("some_branch")
        tracker = bugtracker.tracker_registry.get_tracker("deb", branch)
        self.assertEqual("http://bugs.debian.org/1234", tracker.get_bug_url("1234"))

    def test_gnome_registered(self):
        branch = self.make_branch("some_branch")
        tracker = bugtracker.tracker_registry.get_tracker("gnome", branch)
        self.assertEqual(
            "http://bugzilla.gnome.org/show_bug.cgi?id=1234",
            tracker.get_bug_url("1234"),
        )

    def test_trac_registered(self):
        """The Trac bug tracker should be registered by default and generate
        Trac bug page URLs when the appropriate configuration is present.
        """
        branch = self.make_branch("some_branch")
        config = branch.get_config()
        config.set_user_option("trac_foo_url", "http://bugs.example.com/trac")
        tracker = bugtracker.tracker_registry.get_tracker("foo", branch)
        self.assertEqual(
            "http://bugs.example.com/trac/ticket/1234", tracker.get_bug_url("1234")
        )

    def test_bugzilla_registered(self):
        """The Bugzilla bug tracker should be registered by default and
        generate Bugzilla bug page URLs when the appropriate configuration is
        present.
        """
        branch = self.make_branch("some_branch")
        config = branch.get_config()
        config.set_user_option("bugzilla_foo_url", "http://bugs.example.com")
        tracker = bugtracker.tracker_registry.get_tracker("foo", branch)
        self.assertEqual(
            "http://bugs.example.com/show_bug.cgi?id=1234", tracker.get_bug_url("1234")
        )

    def test_github(self):
        branch = self.make_branch("some_branch")
        tracker = bugtracker.tracker_registry.get_tracker("github", branch)
        self.assertEqual(
            "https://github.com/breezy-team/breezy/issues/1234",
            tracker.get_bug_url("breezy-team/breezy/1234"),
        )

    def test_generic_registered(self):
        branch = self.make_branch("some_branch")
        config = branch.get_config()
        config.set_user_option(
            "bugtracker_foo_url", "http://bugs.example.com/{id}/view.html"
        )
        tracker = bugtracker.tracker_registry.get_tracker("foo", branch)
        self.assertEqual(
            "http://bugs.example.com/1234/view.html", tracker.get_bug_url("1234")
        )

    def test_generic_registered_non_integer(self):
        branch = self.make_branch("some_branch")
        config = branch.get_config()
        config.set_user_option(
            "bugtracker_foo_url", "http://bugs.example.com/{id}/view.html"
        )
        tracker = bugtracker.tracker_registry.get_tracker("foo", branch)
        self.assertEqual(
            "http://bugs.example.com/ABC-1234/view.html",
            tracker.get_bug_url("ABC-1234"),
        )

    def test_generic_incorrect_url(self):
        branch = self.make_branch("some_branch")
        config = branch.get_config()
        config.set_user_option(
            "bugtracker_foo_url", "http://bugs.example.com/view.html"
        )
        tracker = bugtracker.tracker_registry.get_tracker("foo", branch)
        self.assertRaises(bugtracker.InvalidBugTrackerURL, tracker.get_bug_url, "1234")


class TestUniqueIntegerBugTracker(TestCaseWithMemoryTransport):
    def test_appends_id_to_base_url(self):
        """The URL of a bug is the base URL joined to the identifier."""
        tracker = bugtracker.UniqueIntegerBugTracker(
            "xxx", "http://bugs.example.com/foo"
        )
        self.assertEqual("http://bugs.example.com/foo1234", tracker.get_bug_url("1234"))

    def test_returns_tracker_if_abbreviation_matches(self):
        """The get() method should return an instance of the tracker if the
        given abbreviation matches the tracker's abbreviated name.
        """
        tracker = bugtracker.UniqueIntegerBugTracker("xxx", "http://bugs.example.com/")
        branch = self.make_branch("some_branch")
        self.assertIs(tracker, tracker.get("xxx", branch))

    def test_returns_none_if_abbreviation_doesnt_match(self):
        """The get() method should return None if the given abbreviated name
        doesn't match the tracker's abbreviation.
        """
        tracker = bugtracker.UniqueIntegerBugTracker("xxx", "http://bugs.example.com/")
        branch = self.make_branch("some_branch")
        self.assertIs(None, tracker.get("yyy", branch))

    def test_doesnt_consult_branch(self):
        """A UniqueIntegerBugTracker shouldn't consult the branch for tracker
        information.
        """
        tracker = bugtracker.UniqueIntegerBugTracker("xxx", "http://bugs.example.com/")
        self.assertIs(tracker, tracker.get("xxx", None))
        self.assertIs(None, tracker.get("yyy", None))

    def test_check_bug_id_only_accepts_integers(self):
        """A UniqueIntegerBugTracker accepts integers as bug IDs."""
        tracker = bugtracker.UniqueIntegerBugTracker("xxx", "http://bugs.example.com/")
        tracker.check_bug_id("1234")

    def test_check_bug_id_doesnt_accept_non_integers(self):
        """A UniqueIntegerBugTracker rejects non-integers as bug IDs."""
        tracker = bugtracker.UniqueIntegerBugTracker("xxx", "http://bugs.example.com/")
        self.assertRaises(
            bugtracker.MalformedBugIdentifier, tracker.check_bug_id, "red"
        )


class TestProjectIntegerBugTracker(TestCaseWithMemoryTransport):
    def test_appends_id_to_base_url(self):
        """The URL of a bug is the base URL joined to the identifier."""
        tracker = bugtracker.ProjectIntegerBugTracker(
            "xxx", "http://bugs.example.com/{project}/{id}"
        )
        self.assertEqual(
            "http://bugs.example.com/foo/1234", tracker.get_bug_url("foo/1234")
        )

    def test_returns_tracker_if_abbreviation_matches(self):
        """The get() method should return an instance of the tracker if the
        given abbreviation matches the tracker's abbreviated name.
        """
        tracker = bugtracker.ProjectIntegerBugTracker(
            "xxx", "http://bugs.example.com/{project}/{id}"
        )
        branch = self.make_branch("some_branch")
        self.assertIs(tracker, tracker.get("xxx", branch))

    def test_returns_none_if_abbreviation_doesnt_match(self):
        """The get() method should return None if the given abbreviated name
        doesn't match the tracker's abbreviation.
        """
        tracker = bugtracker.ProjectIntegerBugTracker(
            "xxx", "http://bugs.example.com/{project}/{id}"
        )
        branch = self.make_branch("some_branch")
        self.assertIs(None, tracker.get("yyy", branch))

    def test_doesnt_consult_branch(self):
        """Shouldn't consult the branch for tracker information."""
        tracker = bugtracker.ProjectIntegerBugTracker(
            "xxx", "http://bugs.example.com/{project}/{id}"
        )
        self.assertIs(tracker, tracker.get("xxx", None))
        self.assertIs(None, tracker.get("yyy", None))

    def test_check_bug_id_only_accepts_project_integers(self):
        """Accepts integers as bug IDs."""
        tracker = bugtracker.ProjectIntegerBugTracker(
            "xxx", "http://bugs.example.com/{project}/{id}"
        )
        tracker.check_bug_id("project/1234")

    def test_check_bug_id_doesnt_accept_non_project_integers(self):
        """Rejects non-integers as bug IDs."""
        tracker = bugtracker.ProjectIntegerBugTracker(
            "xxx", "http://bugs.example.com/{project}/{id}"
        )
        self.assertRaises(
            bugtracker.MalformedBugIdentifier, tracker.check_bug_id, "red"
        )
        self.assertRaises(
            bugtracker.MalformedBugIdentifier, tracker.check_bug_id, "1234"
        )


class TestURLParametrizedBugTracker(TestCaseWithMemoryTransport):
    """Tests for URLParametrizedBugTracker."""

    def setUp(self):
        super().setUp()
        self.url = "http://twistedmatrix.com/trac"
        self.tracker = bugtracker.URLParametrizedBugTracker("some", "ticket/")

    def test_get_with_unsupported_tag(self):
        """If asked for an unrecognized or unconfigured tag, return None."""
        branch = self.make_branch("some_branch")
        self.assertEqual(None, self.tracker.get("lp", branch))
        self.assertEqual(None, self.tracker.get("twisted", branch))

    def test_get_with_supported_tag(self):
        """If asked for a valid tag, return a tracker instance that can map bug
        IDs to <base_url>/<bug_area> + <bug_id>.
        """
        bugtracker.tracker_registry.register("some", self.tracker)
        self.addCleanup(bugtracker.tracker_registry.remove, "some")

        branch = self.make_branch("some_branch")
        config = branch.get_config()
        config.set_user_option("some_twisted_url", self.url)
        tracker = self.tracker.get("twisted", branch)
        self.assertEqual(
            urlutils.join(self.url, "ticket/") + "1234", tracker.get_bug_url("1234")
        )

    def test_get_bug_url_for_integer_id(self):
        self.tracker.check_bug_id("1234")

    def test_get_bug_url_for_non_integer_id(self):
        self.tracker.check_bug_id("ABC-1234")


class TestURLParametrizedIntegerBugTracker(TestCaseWithMemoryTransport):
    """Tests for URLParametrizedIntegerBugTracker."""

    def setUp(self):
        super().setUp()
        self.url = "http://twistedmatrix.com/trac"
        self.tracker = bugtracker.URLParametrizedIntegerBugTracker("some", "ticket/")

    def test_get_bug_url_for_bad_bug(self):
        """When given a bug identifier that is invalid for Trac, get_bug_url
        should raise an error.
        """
        self.assertRaises(
            bugtracker.MalformedBugIdentifier, self.tracker.get_bug_url, "bad"
        )


class TestPropertyEncoding(TestCase):
    """Tests for how the bug URLs are encoded as revision properties."""

    def test_encoding_one(self):
        self.assertEqual(
            "http://example.com/bugs/1 fixed",
            bugtracker.encode_fixes_bug_urls([("http://example.com/bugs/1", "fixed")]),
        )

    def test_encoding_zero(self):
        self.assertEqual("", bugtracker.encode_fixes_bug_urls([]))

    def test_encoding_two(self):
        self.assertEqual(
            "http://example.com/bugs/1 fixed\nhttp://example.com/bugs/2 related",
            bugtracker.encode_fixes_bug_urls(
                [
                    ("http://example.com/bugs/1", "fixed"),
                    ("http://example.com/bugs/2", "related"),
                ]
            ),
        )

    def test_encoding_with_space(self):
        self.assertRaises(
            bugtracker.InvalidBugUrl,
            bugtracker.encode_fixes_bug_urls,
            [("http://example.com/bugs/ 1", "fixed")],
        )


class TestPropertyDecoding(TestCase):
    """Tests for parsing bug revision properties."""

    def test_decoding_one(self):
        self.assertEqual(
            [("http://example.com/bugs/1", "fixed")],
            list(bugtracker.decode_bug_urls("http://example.com/bugs/1 fixed")),
        )

    def test_decoding_zero(self):
        self.assertEqual([], list(bugtracker.decode_bug_urls("")))

    def test_decoding_two(self):
        self.assertEqual(
            [
                ("http://example.com/bugs/1", "fixed"),
                ("http://example.com/bugs/2", "related"),
            ],
            list(
                bugtracker.decode_bug_urls(
                    "http://example.com/bugs/1 fixed\nhttp://example.com/bugs/2 related"
                )
            ),
        )

    def test_decoding_invalid(self):
        self.assertRaises(
            bugtracker.InvalidLineInBugsProperty,
            list,
            bugtracker.decode_bug_urls("http://example.com/bugs/ 1 fixed\n"),
        )



# Copyright (C) 2005-2012, 2014, 2016, 2017 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import contextlib
import os
import re
import subprocess
import sys
import tempfile
from io import BytesIO

from .. import diff, errors, osutils, revisionspec, revisiontree, tests
from .. import revision as _mod_revision
from ..tests import EncodingAdapter, features
from ..tests.scenarios import load_tests_apply_scenarios

load_tests = load_tests_apply_scenarios


def subst_dates(string):
    """Replace date strings with constant values."""
    return re.sub(
        rb"\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2} [-\+]\d{4}",
        b"YYYY-MM-DD HH:MM:SS +ZZZZ",
        string,
    )


def udiff_lines(old, new, allow_binary=False):
    output = BytesIO()
    diff.internal_diff("old", old, "new", new, output, allow_binary)
    output.seek(0, 0)
    return output.readlines()


def external_udiff_lines(old, new, use_stringio=False):
    if use_stringio:
        # BytesIO has no fileno, so it tests a different codepath
        output = BytesIO()
    else:
        output = tempfile.TemporaryFile()
    try:
        diff.external_diff("old", old, "new", new, output, diff_opts=["-u"])
    except errors.NoDiff:
        raise tests.TestSkipped('external "diff" not present to test')
    output.seek(0, 0)
    lines = output.readlines()
    output.close()
    return lines


class StubO:
    """Simple file-like object that allows writes with any type and records."""

    def __init__(self):
        self.write_record = []

    def write(self, data):
        self.write_record.append(data)

    def check_types(self, testcase, expected_type):
        testcase.assertFalse(
            any(not isinstance(o, expected_type) for o in self.write_record),
            "Not all writes of type {}: {!r}".format(
                expected_type.__name__, self.write_record
            ),
        )


class TestDiffOptions(tests.TestCase):
    def test_unified_added(self):
        """Check for default style '-u' only if no other style specified
        in 'diff-options'.
        """
        # Verify that style defaults to unified, id est '-u' appended
        # to option list, in the absence of an alternative style.
        self.assertEqual(["-a", "-u"], diff.default_style_unified(["-a"]))


class TestDiffOptionsScenarios(tests.TestCase):
    scenarios = [(s, {"style": s}) for s in diff.style_option_list]
    style = None  # Set by load_tests_apply_scenarios from scenarios

    def test_unified_not_added(self):
        # Verify that for all valid style options, '-u' is not
        # appended to option list.
        ret_opts = diff.default_style_unified(diff_opts=["{}".format(self.style)])
        self.assertEqual(["{}".format(self.style)], ret_opts)


class TestDiff(tests.TestCase):
    def test_add_nl(self):
        """Diff generates a valid diff for patches that add a newline."""
        lines = udiff_lines([b"boo"], [b"boo\n"])
        self.check_patch(lines)
        self.assertEqual(lines[4], b"\\ No newline at end of file\n")
        ## "expected no-nl, got %r" % lines[4]

    def test_add_nl_2(self):
        """Diff generates a valid diff for patches that change last line and
        add a newline.
        """
        lines = udiff_lines([b"boo"], [b"goo\n"])
        self.check_patch(lines)
        self.assertEqual(lines[4], b"\\ No newline at end of file\n")
        ## "expected no-nl, got %r" % lines[4]

    def test_remove_nl(self):
        """Diff generates a valid diff for patches that change last line and
        add a newline.
        """
        lines = udiff_lines([b"boo\n"], [b"boo"])
        self.check_patch(lines)
        self.assertEqual(lines[5], b"\\ No newline at end of file\n")
        ## "expected no-nl, got %r" % lines[5]

    def check_patch(self, lines):
        self.assertTrue(len(lines) > 1)
        ## "Not enough lines for a file header for patch:\n%s" % "".join(lines)
        self.assertTrue(lines[0].startswith(b"---"))
        ## 'No orig line for patch:\n%s' % "".join(lines)
        self.assertTrue(lines[1].startswith(b"+++"))
        ## 'No mod line for patch:\n%s' % "".join(lines)
        self.assertTrue(len(lines) > 2)
        ## "No hunks for patch:\n%s" % "".join(lines)
        self.assertTrue(lines[2].startswith(b"@@"))
        ## "No hunk header for patch:\n%s" % "".join(lines)
        self.assertTrue(b"@@" in lines[2][2:])
        ## "Unterminated hunk header for patch:\n%s" % "".join(lines)

    def test_binary_lines(self):
        empty = []
        uni_lines = [1023 * b"a" + b"\x00"]
        self.assertRaises(errors.BinaryFile, udiff_lines, uni_lines, empty)
        self.assertRaises(errors.BinaryFile, udiff_lines, empty, uni_lines)
        udiff_lines(uni_lines, empty, allow_binary=True)
        udiff_lines(empty, uni_lines, allow_binary=True)

    def test_external_diff(self):
        lines = external_udiff_lines([b"boo\n"], [b"goo\n"])
        self.check_patch(lines)
        self.assertEqual(b"\n", lines[-1])

    def test_external_diff_no_fileno(self):
        # Make sure that we can handle not having a fileno, even
        # if the diff is large
        lines = external_udiff_lines(
            [b"boo\n"] * 10000, [b"goo\n"] * 10000, use_stringio=True
        )
        self.check_patch(lines)

    def test_external_diff_binary_lang_c(self):
        for lang in ("LANG", "LC_ALL", "LANGUAGE"):
            self.overrideEnv(lang, "C")
        lines = external_udiff_lines([b"\x00foobar\n"], [b"foo\x00bar\n"])
        # Older versions of diffutils say "Binary files", newer
        # versions just say "Files".
        self.assertContainsRe(lines[0], b"(Binary f|F)iles old and new differ\n")
        self.assertEqual(lines[1:], [b"\n"])

    def test_no_external_diff(self):
        """Check that NoDiff is raised when diff is not available."""
        # Make sure no 'diff' command is available
        # XXX: Weird, using None instead of '' breaks the test -- vila 20101216
        self.overrideEnv("PATH", "")
        self.assertRaises(
            errors.NoDiff,
            diff.external_diff,
            b"old",
            [b"boo\n"],
            b"new",
            [b"goo\n"],
            BytesIO(),
            diff_opts=["-u"],
        )

    def test_internal_diff_default(self):
        # Default internal diff encoding is utf8
        output = BytesIO()
        diff.internal_diff(
            "old_\xb5", [b"old_text\n"], "new_\xe5", [b"new_text\n"], output
        )
        lines = output.getvalue().splitlines(True)
        self.check_patch(lines)
        self.assertEqual(
            [
                b"--- old_\xc2\xb5\n",
                b"+++ new_\xc3\xa5\n",
                b"@@ -1,1 +1,1 @@\n",
                b"-old_text\n",
                b"+new_text\n",
                b"\n",
            ],
            lines,
        )

    def test_internal_diff_utf8(self):
        output = BytesIO()
        diff.internal_diff(
            "old_\xb5",
            [b"old_text\n"],
            "new_\xe5",
            [b"new_text\n"],
            output,
            path_encoding="utf8",
        )
        lines = output.getvalue().splitlines(True)
        self.check_patch(lines)
        self.assertEqual(
            [
                b"--- old_\xc2\xb5\n",
                b"+++ new_\xc3\xa5\n",
                b"@@ -1,1 +1,1 @@\n",
                b"-old_text\n",
                b"+new_text\n",
                b"\n",
            ],
            lines,
        )

    def test_internal_diff_iso_8859_1(self):
        output = BytesIO()
        diff.internal_diff(
            "old_\xb5",
            [b"old_text\n"],
            "new_\xe5",
            [b"new_text\n"],
            output,
            path_encoding="iso-8859-1",
        )
        lines = output.getvalue().splitlines(True)
        self.check_patch(lines)
        self.assertEqual(
            [
                b"--- old_\xb5\n",
                b"+++ new_\xe5\n",
                b"@@ -1,1 +1,1 @@\n",
                b"-old_text\n",
                b"+new_text\n",
                b"\n",
            ],
            lines,
        )

    def test_internal_diff_no_content(self):
        output = BytesIO()
        diff.internal_diff("old", [], "new", [], output)
        self.assertEqual(b"", output.getvalue())

    def test_internal_diff_no_changes(self):
        output = BytesIO()
        diff.internal_diff(
            "old", [b"text\n", b"contents\n"], "new", [b"text\n", b"contents\n"], output
        )
        self.assertEqual(b"", output.getvalue())

    def test_internal_diff_returns_bytes(self):
        output = StubO()
        diff.internal_diff(
            "old_\xb5", [b"old_text\n"], "new_\xe5", [b"new_text\n"], output
        )
        output.check_types(self, bytes)

    def test_internal_diff_default_context(self):
        output = BytesIO()
        diff.internal_diff(
            "old",
            [
                b"same_text\n",
                b"same_text\n",
                b"same_text\n",
                b"same_text\n",
                b"same_text\n",
                b"old_text\n",
            ],
            "new",
            [
                b"same_text\n",
                b"same_text\n",
                b"same_text\n",
                b"same_text\n",
                b"same_text\n",
                b"new_text\n",
            ],
            output,
        )
        lines = output.getvalue().splitlines(True)
        self.check_patch(lines)
        self.assertEqual(
            [
                b"--- old\n",
                b"+++ new\n",
                b"@@ -3,4 +3,4 @@\n",
                b" same_text\n",
                b" same_text\n",
                b" same_text\n",
                b"-old_text\n",
                b"+new_text\n",
                b"\n",
            ],
            lines,
        )

    def test_internal_diff_no_context(self):
        output = BytesIO()
        diff.internal_diff(
            "old",
            [
                b"same_text\n",
                b"same_text\n",
                b"same_text\n",
                b"same_text\n",
                b"same_text\n",
                b"old_text\n",
            ],
            "new",
            [
                b"same_text\n",
                b"same_text\n",
                b"same_text\n",
                b"same_text\n",
                b"same_text\n",
                b"new_text\n",
            ],
            output,
            context_lines=0,
        )
        lines = output.getvalue().splitlines(True)
        self.check_patch(lines)
        self.assertEqual(
            [
                b"--- old\n",
                b"+++ new\n",
                b"@@ -6,1 +6,1 @@\n",
                b"-old_text\n",
                b"+new_text\n",
                b"\n",
            ],
            lines,
        )

    def test_internal_diff_more_context(self):
        output = BytesIO()
        diff.internal_diff(
            "old",
            [
                b"same_text\n",
                b"same_text\n",
                b"same_text\n",
                b"same_text\n",
                b"same_text\n",
                b"old_text\n",
            ],
            "new",
            [
                b"same_text\n",
                b"same_text\n",
                b"same_text\n",
                b"same_text\n",
                b"same_text\n",
                b"new_text\n",
            ],
            output,
            context_lines=4,
        )
        lines = output.getvalue().splitlines(True)
        self.check_patch(lines)
        self.assertEqual(
            [
                b"--- old\n",
                b"+++ new\n",
                b"@@ -2,5 +2,5 @@\n",
                b" same_text\n",
                b" same_text\n",
                b" same_text\n",
                b" same_text\n",
                b"-old_text\n",
                b"+new_text\n",
                b"\n",
            ],
            lines,
        )


class TestDiffFiles(tests.TestCaseInTempDir):
    def test_external_diff_binary(self):
        """The output when using external diff should use diff's i18n error."""
        for lang in ("LANG", "LC_ALL", "LANGUAGE"):
            self.overrideEnv(lang, "C")
        # Make sure external_diff doesn't fail in the current LANG
        lines = external_udiff_lines([b"\x00foobar\n"], [b"foo\x00bar\n"])

        cmd = ["diff", "-u", "--binary", "old", "new"]
        with open("old", "wb") as f:
            f.write(b"\x00foobar\n")
        with open("new", "wb") as f:
            f.write(b"foo\x00bar\n")
        pipe = subprocess.Popen(cmd, stdout=subprocess.PIPE, stdin=subprocess.PIPE)
        out, err = pipe.communicate()
        # We should output whatever diff tells us, plus a trailing newline
        self.assertEqual(out.splitlines(True) + [b"\n"], lines)


def get_diff_as_string(tree1, tree2, specific_files=None, working_tree=None):
    output = BytesIO()
    if working_tree is not None:
        extra_trees = (working_tree,)
    else:
        extra_trees = ()
    diff.show_diff_trees(
        tree1,
        tree2,
        output,
        specific_files=specific_files,
        extra_trees=extra_trees,
        old_label="old/",
        new_label="new/",
    )
    return output.getvalue()


class TestDiffDates(tests.TestCaseWithTransport):
    def setUp(self):
        super().setUp()
        self.wt = self.make_branch_and_tree(".")
        self.b = self.wt.branch
        self.build_tree_contents(
            [
                ("file1", b"file1 contents at rev 1\n"),
                ("file2", b"file2 contents at rev 1\n"),
            ]
        )
        self.wt.add(["file1", "file2"])
        self.wt.commit(
            message="Revision 1",
            timestamp=1143849600,  # 2006-04-01 00:00:00 UTC
            timezone=0,
            rev_id=b"rev-1",
        )
        self.build_tree_contents([("file1", b"file1 contents at rev 2\n")])
        self.wt.commit(
            message="Revision 2",
            timestamp=1143936000,  # 2006-04-02 00:00:00 UTC
            timezone=28800,
            rev_id=b"rev-2",
        )
        self.build_tree_contents([("file2", b"file2 contents at rev 3\n")])
        self.wt.commit(
            message="Revision 3",
            timestamp=1144022400,  # 2006-04-03 00:00:00 UTC
            timezone=-3600,
            rev_id=b"rev-3",
        )
        self.wt.remove(["file2"])
        self.wt.commit(
            message="Revision 4",
            timestamp=1144108800,  # 2006-04-04 00:00:00 UTC
            timezone=0,
            rev_id=b"rev-4",
        )
        self.build_tree_contents([("file1", b"file1 contents in working tree\n")])
        # set the date stamps for files in the working tree to known values
        os.utime("file1", (1144195200, 1144195200))  # 2006-04-05 00:00:00 UTC

    def test_diff_rev_tree_working_tree(self):
        output = get_diff_as_string(self.wt.basis_tree(), self.wt)
        # note that the date for old/file1 is from rev 2 rather than from
        # the basis revision (rev 4)
        self.assertEqualDiff(
            output,
            b"""\
=== modified file 'file1'
--- old/file1\t2006-04-02 00:00:00 +0000
+++ new/file1\t2006-04-05 00:00:00 +0000
@@ -1,1 +1,1 @@
-file1 contents at rev 2
+file1 contents in working tree

""",
        )

    def test_diff_rev_tree_rev_tree(self):
        tree1 = self.b.repository.revision_tree(b"rev-2")
        tree2 = self.b.repository.revision_tree(b"rev-3")
        output = get_diff_as_string(tree1, tree2)
        self.assertEqualDiff(
            output,
            b"""\
=== modified file 'file2'
--- old/file2\t2006-04-01 00:00:00 +0000
+++ new/file2\t2006-04-03 00:00:00 +0000
@@ -1,1 +1,1 @@
-file2 contents at rev 1
+file2 contents at rev 3

""",
        )

    def test_diff_add_files(self):
        tree1 = self.b.repository.revision_tree(_mod_revision.NULL_REVISION)
        tree2 = self.b.repository.revision_tree(b"rev-1")
        output = get_diff_as_string(tree1, tree2)
        # the files have the epoch time stamp for the tree in which
        # they don't exist.
        self.assertEqualDiff(
            output,
            b"""\
=== added file 'file1'
--- old/file1\t1970-01-01 00:00:00 +0000
+++ new/file1\t2006-04-01 00:00:00 +0000
@@ -0,0 +1,1 @@
+file1 contents at rev 1

=== added file 'file2'
--- old/file2\t1970-01-01 00:00:00 +0000
+++ new/file2\t2006-04-01 00:00:00 +0000
@@ -0,0 +1,1 @@
+file2 contents at rev 1

""",
        )

    def test_diff_remove_files(self):
        tree1 = self.b.repository.revision_tree(b"rev-3")
        tree2 = self.b.repository.revision_tree(b"rev-4")
        output = get_diff_as_string(tree1, tree2)
        # the file has the epoch time stamp for the tree in which
        # it doesn't exist.
        self.assertEqualDiff(
            output,
            b"""\
=== removed file 'file2'
--- old/file2\t2006-04-03 00:00:00 +0000
+++ new/file2\t1970-01-01 00:00:00 +0000
@@ -1,1 +0,0 @@
-file2 contents at rev 3

""",
        )

    def test_show_diff_specified(self):
        """A working tree filename can be used to identify a file."""
        self.wt.rename_one("file1", "file1b")
        old_tree = self.b.repository.revision_tree(b"rev-1")
        new_tree = self.b.repository.revision_tree(b"rev-4")
        out = get_diff_as_string(
            old_tree, new_tree, specific_files=["file1b"], working_tree=self.wt
        )
        self.assertContainsRe(out, b"file1\t")

    def test_recursive_diff(self):
        """Children of directories are matched."""
        os.mkdir("dir1")
        os.mkdir("dir2")
        self.wt.add(["dir1", "dir2"])
        self.wt.rename_one("file1", "dir1/file1")
        old_tree = self.b.repository.revision_tree(b"rev-1")
        new_tree = self.b.repository.revision_tree(b"rev-4")
        out = get_diff_as_string(
            old_tree, new_tree, specific_files=["dir1"], working_tree=self.wt
        )
        self.assertContainsRe(out, b"file1\t")
        out = get_diff_as_string(
            old_tree, new_tree, specific_files=["dir2"], working_tree=self.wt
        )
        self.assertNotContainsRe(out, b"file1\t")


class TestShowDiffTrees(tests.TestCaseWithTransport):
    """Direct tests for show_diff_trees."""

    def test_modified_file(self):
        """Test when a file is modified."""
        tree = self.make_branch_and_tree("tree")
        self.build_tree_contents([("tree/file", b"contents\n")])
        tree.add(["file"], ids=[b"file-id"])
        tree.commit("one", rev_id=b"rev-1")

        self.build_tree_contents([("tree/file", b"new contents\n")])
        d = get_diff_as_string(tree.basis_tree(), tree)
        self.assertContainsRe(d, b"=== modified file 'file'\n")
        self.assertContainsRe(d, b"--- old/file\t")
        self.assertContainsRe(d, b"\\+\\+\\+ new/file\t")
        self.assertContainsRe(d, b"-contents\n\\+new contents\n")

    def test_modified_file_in_renamed_dir(self):
        """Test when a file is modified in a renamed directory."""
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/dir/"])
        self.build_tree_contents([("tree/dir/file", b"contents\n")])
        tree.add(["dir", "dir/file"], ids=[b"dir-id", b"file-id"])
        tree.commit("one", rev_id=b"rev-1")

        tree.rename_one("dir", "other")
        self.build_tree_contents([("tree/other/file", b"new contents\n")])
        d = get_diff_as_string(tree.basis_tree(), tree)
        self.assertContainsRe(d, b"=== renamed directory 'dir' => 'other'\n")
        self.assertContainsRe(d, b"=== modified file 'other/file'\n")
        # XXX: This is technically incorrect, because it used to be at another
        # location. What to do?
        self.assertContainsRe(d, b"--- old/dir/file\t")
        self.assertContainsRe(d, b"\\+\\+\\+ new/other/file\t")
        self.assertContainsRe(d, b"-contents\n\\+new contents\n")

    def test_renamed_directory(self):
        """Test when only a directory is only renamed."""
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/dir/"])
        self.build_tree_contents([("tree/dir/file", b"contents\n")])
        tree.add(["dir", "dir/file"], ids=[b"dir-id", b"file-id"])
        tree.commit("one", rev_id=b"rev-1")

        tree.rename_one("dir", "newdir")
        d = get_diff_as_string(tree.basis_tree(), tree)
        # Renaming a directory should be a single "you renamed this dir" even
        # when there are files inside.
        self.assertEqual(d, b"=== renamed directory 'dir' => 'newdir'\n")

    def test_renamed_file(self):
        """Test when a file is only renamed."""
        tree = self.make_branch_and_tree("tree")
        self.build_tree_contents([("tree/file", b"contents\n")])
        tree.add(["file"], ids=[b"file-id"])
        tree.commit("one", rev_id=b"rev-1")

        tree.rename_one("file", "newname")
        d = get_diff_as_string(tree.basis_tree(), tree)
        self.assertContainsRe(d, b"=== renamed file 'file' => 'newname'\n")
        # We shouldn't have a --- or +++ line, because there is no content
        # change
        self.assertNotContainsRe(d, b"---")

    def test_renamed_and_modified_file(self):
        """Test when a file is only renamed."""
        tree = self.make_branch_and_tree("tree")
        self.build_tree_contents([("tree/file", b"contents\n")])
        tree.add(["file"], ids=[b"file-id"])
        tree.commit("one", rev_id=b"rev-1")

        tree.rename_one("file", "newname")
        self.build_tree_contents([("tree/newname", b"new contents\n")])
        d = get_diff_as_string(tree.basis_tree(), tree)
        self.assertContainsRe(d, b"=== renamed file 'file' => 'newname'\n")
        self.assertContainsRe(d, b"--- old/file\t")
        self.assertContainsRe(d, b"\\+\\+\\+ new/newname\t")
        self.assertContainsRe(d, b"-contents\n\\+new contents\n")

    def test_internal_diff_exec_property(self):
        tree = self.make_branch_and_tree("tree")

        tt = tree.transform()
        tt.new_file("a", tt.root, [b"contents\n"], b"a-id", True)
        tt.new_file("b", tt.root, [b"contents\n"], b"b-id", False)
        tt.new_file("c", tt.root, [b"contents\n"], b"c-id", True)
        tt.new_file("d", tt.root, [b"contents\n"], b"d-id", False)
        tt.new_file("e", tt.root, [b"contents\n"], b"control-e-id", True)
        tt.new_file("f", tt.root, [b"contents\n"], b"control-f-id", False)
        tt.apply()
        tree.commit("one", rev_id=b"rev-1")

        tt = tree.transform()
        tt.set_executability(False, tt.trans_id_file_id(b"a-id"))
        tt.set_executability(True, tt.trans_id_file_id(b"b-id"))
        tt.set_executability(False, tt.trans_id_file_id(b"c-id"))
        tt.set_executability(True, tt.trans_id_file_id(b"d-id"))
        tt.apply()
        tree.rename_one("c", "new-c")
        tree.rename_one("d", "new-d")

        d = get_diff_as_string(tree.basis_tree(), tree)

        self.assertContainsRe(
            d, rb"file 'a'.*\(properties changed:" rb".*\+x to -x.*\)"
        )
        self.assertContainsRe(
            d, rb"file 'b'.*\(properties changed:" rb".*-x to \+x.*\)"
        )
        self.assertContainsRe(
            d, rb"file 'c'.*\(properties changed:" rb".*\+x to -x.*\)"
        )
        self.assertContainsRe(
            d, rb"file 'd'.*\(properties changed:" rb".*-x to \+x.*\)"
        )
        self.assertNotContainsRe(d, rb"file 'e'")
        self.assertNotContainsRe(d, rb"file 'f'")

    def test_binary_unicode_filenames(self):
        """Test that contents of files are *not* encoded in UTF-8 when there
        is a binary file in the diff.
        """
        # See https://bugs.launchpad.net/bugs/110092.
        self.requireFeature(features.UnicodeFilenameFeature)

        tree = self.make_branch_and_tree("tree")
        alpha, omega = "\u03b1", "\u03c9"
        alpha_utf8, omega_utf8 = alpha.encode("utf8"), omega.encode("utf8")
        self.build_tree_contents(
            [
                ("tree/" + alpha, b"\0"),
                ("tree/" + omega, (b"The %s and the %s\n" % (alpha_utf8, omega_utf8))),
            ]
        )
        tree.add([alpha])
        tree.add([omega])
        diff_content = StubO()
        diff.show_diff_trees(tree.basis_tree(), tree, diff_content)
        diff_content.check_types(self, bytes)
        d = b"".join(diff_content.write_record)
        self.assertContainsRe(d, rb"=== added file '%s'" % alpha_utf8)
        self.assertContainsRe(
            d, b"Binary files a/%s.*and b/%s.* differ\n" % (alpha_utf8, alpha_utf8)
        )
        self.assertContainsRe(d, rb"=== added file '%s'" % omega_utf8)
        self.assertContainsRe(d, rb"--- a/%s" % (omega_utf8,))
        self.assertContainsRe(d, rb"\+\+\+ b/%s" % (omega_utf8,))

    def test_unicode_filename(self):
        """Test when the filename are unicode."""
        self.requireFeature(features.UnicodeFilenameFeature)

        alpha, omega = "\u03b1", "\u03c9"
        autf8, outf8 = alpha.encode("utf8"), omega.encode("utf8")

        tree = self.make_branch_and_tree("tree")
        self.build_tree_contents([("tree/ren_" + alpha, b"contents\n")])
        tree.add(["ren_" + alpha], ids=[b"file-id-2"])
        self.build_tree_contents([("tree/del_" + alpha, b"contents\n")])
        tree.add(["del_" + alpha], ids=[b"file-id-3"])
        self.build_tree_contents([("tree/mod_" + alpha, b"contents\n")])
        tree.add(["mod_" + alpha], ids=[b"file-id-4"])

        tree.commit("one", rev_id=b"rev-1")

        tree.rename_one("ren_" + alpha, "ren_" + omega)
        tree.remove("del_" + alpha)
        self.build_tree_contents([("tree/add_" + alpha, b"contents\n")])
        tree.add(["add_" + alpha], ids=[b"file-id"])
        self.build_tree_contents([("tree/mod_" + alpha, b"contents_mod\n")])

        d = get_diff_as_string(tree.basis_tree(), tree)
        self.assertContainsRe(
            d, b"=== renamed file 'ren_%s' => 'ren_%s'\n" % (autf8, outf8)
        )
        self.assertContainsRe(d, b"=== added file 'add_%s'" % autf8)
        self.assertContainsRe(d, b"=== modified file 'mod_%s'" % autf8)
        self.assertContainsRe(d, b"=== removed file 'del_%s'" % autf8)

    def test_unicode_filename_path_encoding(self):
        """Test for bug #382699: unicode filenames on Windows should be shown
        in user encoding.
        """
        self.requireFeature(features.UnicodeFilenameFeature)
        # The word 'test' in Russian
        _russian_test = "\u0422\u0435\u0441\u0442"
        directory = _russian_test + "/"
        test_txt = _russian_test + ".txt"
        u1234 = "\u1234.txt"

        tree = self.make_branch_and_tree(".")
        self.build_tree_contents(
            [
                (test_txt, b"foo\n"),
                (u1234, b"foo\n"),
                (directory, None),
            ]
        )
        tree.add([test_txt, u1234, directory])

        sio = BytesIO()
        diff.show_diff_trees(tree.basis_tree(), tree, sio, path_encoding="cp1251")

        output = subst_dates(sio.getvalue())
        shouldbe = b"""\
=== added directory '%(directory)s'
=== added file '%(test_txt)s'
--- a/%(test_txt)s\tYYYY-MM-DD HH:MM:SS +ZZZZ
+++ b/%(test_txt)s\tYYYY-MM-DD HH:MM:SS +ZZZZ
@@ -0,0 +1,1 @@
+foo

=== added file '?.txt'
--- a/?.txt\tYYYY-MM-DD HH:MM:SS +ZZZZ
+++ b/?.txt\tYYYY-MM-DD HH:MM:SS +ZZZZ
@@ -0,0 +1,1 @@
+foo

""" % {
            b"directory": _russian_test.encode("cp1251"),
            b"test_txt": test_txt.encode("cp1251"),
        }
        self.assertEqualDiff(output, shouldbe)


class DiffWasIs(diff.DiffPath):
    def diff(self, old_path, new_path, old_kind, new_kind):
        self.to_file.write(b"was: ")
        self.to_file.write(self.old_tree.get_file(old_path).read())
        self.to_file.write(b"is: ")
        self.to_file.write(self.new_tree.get_file(new_path).read())


class TestDiffTree(tests.TestCaseWithTransport):
    def setUp(self):
        super().setUp()
        self.old_tree = self.make_branch_and_tree("old-tree")
        self.old_tree.lock_write()
        self.addCleanup(self.old_tree.unlock)
        self.new_tree = self.make_branch_and_tree("new-tree")
        self.new_tree.lock_write()
        self.addCleanup(self.new_tree.unlock)
        self.differ = diff.DiffTree(self.old_tree, self.new_tree, BytesIO())

    def test_diff_text(self):
        self.build_tree_contents(
            [("old-tree/olddir/",), ("old-tree/olddir/oldfile", b"old\n")]
        )
        self.old_tree.add("olddir")
        self.old_tree.add("olddir/oldfile", ids=b"file-id")
        self.build_tree_contents(
            [("new-tree/newdir/",), ("new-tree/newdir/newfile", b"new\n")]
        )
        self.new_tree.add("newdir")
        self.new_tree.add("newdir/newfile", ids=b"file-id")
        differ = diff.DiffText(self.old_tree, self.new_tree, BytesIO())
        differ.diff_text("olddir/oldfile", None, "old label", "new label")
        self.assertEqual(
            b"--- old label\n+++ new label\n@@ -1,1 +0,0 @@\n-old\n\n",
            differ.to_file.getvalue(),
        )
        differ.to_file.seek(0)
        differ.diff_text(None, "newdir/newfile", "old label", "new label")
        self.assertEqual(
            b"--- old label\n+++ new label\n@@ -0,0 +1,1 @@\n+new\n\n",
            differ.to_file.getvalue(),
        )
        differ.to_file.seek(0)
        differ.diff_text("olddir/oldfile", "newdir/newfile", "old label", "new label")
        self.assertEqual(
            b"--- old label\n+++ new label\n@@ -1,1 +1,1 @@\n-old\n+new\n\n",
            differ.to_file.getvalue(),
        )

    def test_diff_deletion(self):
        self.build_tree_contents(
            [("old-tree/file", b"contents"), ("new-tree/file", b"contents")]
        )
        self.old_tree.add("file", ids=b"file-id")
        self.new_tree.add("file", ids=b"file-id")
        os.unlink("new-tree/file")
        self.differ.show_diff(None)
        self.assertContainsRe(self.differ.to_file.getvalue(), b"-contents")

    def test_diff_creation(self):
        self.build_tree_contents(
            [("old-tree/file", b"contents"), ("new-tree/file", b"contents")]
        )
        self.old_tree.add("file", ids=b"file-id")
        self.new_tree.add("file", ids=b"file-id")
        os.unlink("old-tree/file")
        self.differ.show_diff(None)
        self.assertContainsRe(self.differ.to_file.getvalue(), rb"\+contents")

    def test_diff_symlink(self):
        differ = diff.DiffSymlink(self.old_tree, self.new_tree, BytesIO())
        differ.diff_symlink("old target", None)
        self.assertEqual(b"=== target was 'old target'\n", differ.to_file.getvalue())

        differ = diff.DiffSymlink(self.old_tree, self.new_tree, BytesIO())
        differ.diff_symlink(None, "new target")
        self.assertEqual(b"=== target is 'new target'\n", differ.to_file.getvalue())

        differ = diff.DiffSymlink(self.old_tree, self.new_tree, BytesIO())
        differ.diff_symlink("old target", "new target")
        self.assertEqual(
            b"=== target changed 'old target' => 'new target'\n",
            differ.to_file.getvalue(),
        )

    def test_diff(self):
        self.build_tree_contents(
            [("old-tree/olddir/",), ("old-tree/olddir/oldfile", b"old\n")]
        )
        self.old_tree.add("olddir")
        self.old_tree.add("olddir/oldfile", ids=b"file-id")
        self.build_tree_contents(
            [("new-tree/newdir/",), ("new-tree/newdir/newfile", b"new\n")]
        )
        self.new_tree.add("newdir")
        self.new_tree.add("newdir/newfile", ids=b"file-id")
        self.differ.diff("olddir/oldfile", "newdir/newfile")
        self.assertContainsRe(
            self.differ.to_file.getvalue(),
            rb"--- olddir/oldfile.*\n\+\+\+ newdir/newfile.*\n\@\@ -1,1 \+1,1"
            rb" \@\@\n-old\n\+new\n\n",
        )

    def test_diff_kind_change(self):
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        self.build_tree_contents(
            [("old-tree/olddir/",), ("old-tree/olddir/oldfile", b"old\n")]
        )
        self.old_tree.add("olddir")
        self.old_tree.add("olddir/oldfile", ids=b"file-id")
        self.build_tree(["new-tree/newdir/"])
        os.symlink("new", "new-tree/newdir/newfile")
        self.new_tree.add("newdir")
        self.new_tree.add("newdir/newfile", ids=b"file-id")
        self.differ.diff("olddir/oldfile", "newdir/newfile")
        self.assertContainsRe(
            self.differ.to_file.getvalue(),
            rb"--- olddir/oldfile.*\n"
            rb"\+\+\+ newdir/newfile.*\n"
            rb"\@\@ -1,1 \+0,0 \@\@\n"
            rb"-old\n"
            rb"\n",
        )
        self.assertContainsRe(self.differ.to_file.getvalue(), b"=== target is 'new'\n")

    def test_diff_directory(self):
        self.build_tree(["new-tree/new-dir/"])
        self.new_tree.add("new-dir", ids=b"new-dir-id")
        self.differ.diff(None, "new-dir")
        self.assertEqual(self.differ.to_file.getvalue(), b"")

    def create_old_new(self):
        self.build_tree_contents(
            [("old-tree/olddir/",), ("old-tree/olddir/oldfile", b"old\n")]
        )
        self.old_tree.add("olddir")
        self.old_tree.add("olddir/oldfile", ids=b"file-id")
        self.build_tree_contents(
            [("new-tree/newdir/",), ("new-tree/newdir/newfile", b"new\n")]
        )
        self.new_tree.add("newdir")
        self.new_tree.add("newdir/newfile", ids=b"file-id")

    def test_register_diff(self):
        self.create_old_new()
        old_diff_factories = diff.DiffTree.diff_factories
        diff.DiffTree.diff_factories = old_diff_factories[:]
        diff.DiffTree.diff_factories.insert(0, DiffWasIs.from_diff_tree)
        try:
            differ = diff.DiffTree(self.old_tree, self.new_tree, BytesIO())
        finally:
            diff.DiffTree.diff_factories = old_diff_factories
        differ.diff("olddir/oldfile", "newdir/newfile")
        self.assertNotContainsRe(
            differ.to_file.getvalue(),
            rb"--- olddir/oldfile.*\n\+\+\+ newdir/newfile.*\n\@\@ -1,1 \+1,1"
            rb" \@\@\n-old\n\+new\n\n",
        )
        self.assertContainsRe(differ.to_file.getvalue(), b"was: old\nis: new\n")

    def test_extra_factories(self):
        self.create_old_new()
        differ = diff.DiffTree(
            self.old_tree,
            self.new_tree,
            BytesIO(),
            extra_factories=[DiffWasIs.from_diff_tree],
        )
        differ.diff("olddir/oldfile", "newdir/newfile")
        self.assertNotContainsRe(
            differ.to_file.getvalue(),
            rb"--- olddir/oldfile.*\n\+\+\+ newdir/newfile.*\n\@\@ -1,1 \+1,1"
            rb" \@\@\n-old\n\+new\n\n",
        )
        self.assertContainsRe(differ.to_file.getvalue(), b"was: old\nis: new\n")

    def test_alphabetical_order(self):
        self.build_tree(["new-tree/a-file"])
        self.new_tree.add("a-file")
        self.build_tree(["old-tree/b-file"])
        self.old_tree.add("b-file")
        self.differ.show_diff(None)
        self.assertContainsRe(self.differ.to_file.getvalue(), b".*a-file(.|\n)*b-file")


class TestDiffFromTool(tests.TestCaseWithTransport):
    def test_from_string(self):
        diff_obj = diff.DiffFromTool.from_string(
            ["diff", "{old_path}", "{new_path}"], None, None, None
        )
        self.addCleanup(diff_obj.finish)
        self.assertEqual(
            ["diff", "{old_path}", "{new_path}"], diff_obj.command_template
        )

    def test_from_string_no_paths(self):
        diff_obj = diff.DiffFromTool.from_string(["diff", "-u5"], None, None, None)
        self.addCleanup(diff_obj.finish)
        self.assertEqual(["diff", "-u5"], diff_obj.command_template)
        self.assertEqual(
            ["diff", "-u5", "old-path", "new-path"],
            diff_obj._get_command("old-path", "new-path"),
        )

    def test_from_string_u5(self):
        diff_obj = diff.DiffFromTool.from_string(
            ["diff", "-u 5", "{old_path}", "{new_path}"], None, None, None
        )
        self.addCleanup(diff_obj.finish)
        self.assertEqual(
            ["diff", "-u 5", "{old_path}", "{new_path}"], diff_obj.command_template
        )
        self.assertEqual(
            ["diff", "-u 5", "old-path", "new-path"],
            diff_obj._get_command("old-path", "new-path"),
        )

    def test_from_string_path_with_backslashes(self):
        self.requireFeature(features.backslashdir_feature)
        tool = ["C:\\Tools\\Diff.exe", "{old_path}", "{new_path}"]
        diff_obj = diff.DiffFromTool.from_string(tool, None, None, None)
        self.addCleanup(diff_obj.finish)
        self.assertEqual(
            ["C:\\Tools\\Diff.exe", "{old_path}", "{new_path}"],
            diff_obj.command_template,
        )
        self.assertEqual(
            ["C:\\Tools\\Diff.exe", "old-path", "new-path"],
            diff_obj._get_command("old-path", "new-path"),
        )

    def test_execute(self):
        output = BytesIO()
        diff_obj = diff.DiffFromTool(
            [sys.executable, "-c", 'print("{old_path} {new_path}")'], None, None, output
        )
        self.addCleanup(diff_obj.finish)
        diff_obj._execute("old", "new")
        self.assertEqual(output.getvalue().rstrip(), b"old new")

    def test_execute_missing(self):
        diff_obj = diff.DiffFromTool(
            ["a-tool-which-is-unlikely-to-exist"], None, None, None
        )
        self.addCleanup(diff_obj.finish)
        e = self.assertRaises(errors.ExecutableMissing, diff_obj._execute, "old", "new")
        self.assertEqual(
            "a-tool-which-is-unlikely-to-exist could not be found on this machine",
            str(e),
        )

    def test_prepare_files_creates_paths_readable_by_windows_tool(self):
        self.requireFeature(features.AttribFeature)
        output = BytesIO()
        tree = self.make_branch_and_tree("tree")
        self.build_tree_contents([("tree/file", b"content")])
        tree.add("file", b"file-id")
        tree.commit("old tree")
        tree.lock_read()
        self.addCleanup(tree.unlock)
        basis_tree = tree.basis_tree()
        basis_tree.lock_read()
        self.addCleanup(basis_tree.unlock)
        diff_obj = diff.DiffFromTool(
            [sys.executable, "-c", 'print "{old_path} {new_path}"'],
            basis_tree,
            tree,
            output,
        )
        diff_obj._prepare_files("file", "file", file_id=b"file-id")
        # The old content should be readonly
        self.assertReadableByAttrib(diff_obj._root, "old\\file", r"R.*old\\file$")
        # The new content should use the tree object, not a 'new' file anymore
        self.assertEndsWith(tree.basedir, "work/tree")
        self.assertReadableByAttrib(tree.basedir, "file", r"work\\tree\\file$")

    def assertReadableByAttrib(self, cwd, relpath, regex):
        proc = subprocess.Popen(["attrib", relpath], stdout=subprocess.PIPE, cwd=cwd)
        (result, err) = proc.communicate()
        self.assertContainsRe(result.replace("\r\n", "\n"), regex)

    def test_prepare_files(self):
        output = BytesIO()
        tree = self.make_branch_and_tree("tree")
        self.build_tree_contents([("tree/oldname", b"oldcontent")])
        self.build_tree_contents([("tree/oldname2", b"oldcontent2")])
        tree.add("oldname", ids=b"file-id")
        tree.add("oldname2", ids=b"file2-id")
        # Earliest allowable date on FAT32 filesystems is 1980-01-01
        tree.commit("old tree", timestamp=315532800)
        tree.rename_one("oldname", "newname")
        tree.rename_one("oldname2", "newname2")
        self.build_tree_contents([("tree/newname", b"newcontent")])
        self.build_tree_contents([("tree/newname2", b"newcontent2")])
        old_tree = tree.basis_tree()
        old_tree.lock_read()
        self.addCleanup(old_tree.unlock)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        diff_obj = diff.DiffFromTool(
            [sys.executable, "-c", 'print "{old_path} {new_path}"'],
            old_tree,
            tree,
            output,
        )
        self.addCleanup(diff_obj.finish)
        self.assertContainsRe(diff_obj._root, "brz-diff-[^/]*")
        old_path, new_path = diff_obj._prepare_files("oldname", "newname")
        self.assertContainsRe(old_path, "old/oldname$")
        self.assertEqual(315532800, os.stat(old_path).st_mtime)
        self.assertContainsRe(new_path, "tree/newname$")
        self.assertFileEqual(b"oldcontent", old_path)
        self.assertFileEqual(b"newcontent", new_path)
        if osutils.supports_symlinks(self.test_dir):
            self.assertTrue(os.path.samefile("tree/newname", new_path))
        # make sure we can create files with the same parent directories
        diff_obj._prepare_files("oldname2", "newname2")


class TestDiffFromToolEncodedFilename(tests.TestCaseWithTransport):
    def test_encodable_filename(self):
        # Just checks file path for external diff tool.
        # We cannot change CPython's internal encoding used by os.exec*.
        diffobj = diff.DiffFromTool(
            ["dummy", "{old_path}", "{new_path}"], None, None, None
        )
        for _, scenario in EncodingAdapter.encoding_scenarios:
            encoding = scenario["encoding"]
            dirname = scenario["info"]["directory"]
            filename = scenario["info"]["filename"]

            self.overrideAttr(diffobj, "_fenc", lambda: encoding)
            relpath = dirname + "/" + filename
            fullpath = diffobj._safe_filename("safe", relpath)
            self.assertEqual(fullpath, fullpath.encode(encoding).decode(encoding))
            self.assertTrue(fullpath.startswith(diffobj._root + "/safe"))

    def test_unencodable_filename(self):
        diffobj = diff.DiffFromTool(
            ["dummy", "{old_path}", "{new_path}"], None, None, None
        )
        for _, scenario in EncodingAdapter.encoding_scenarios:
            encoding = scenario["encoding"]
            dirname = scenario["info"]["directory"]
            filename = scenario["info"]["filename"]

            if encoding == "iso-8859-1":
                encoding = "iso-8859-2"
            else:
                encoding = "iso-8859-1"

            self.overrideAttr(diffobj, "_fenc", lambda: encoding)
            relpath = dirname + "/" + filename
            fullpath = diffobj._safe_filename("safe", relpath)
            self.assertEqual(fullpath, fullpath.encode(encoding).decode(encoding))
            self.assertTrue(fullpath.startswith(diffobj._root + "/safe"))


class TestGetTreesAndBranchesToDiffLocked(tests.TestCaseWithTransport):
    def call_gtabtd(self, path_list, revision_specs, old_url, new_url):
        """Call get_trees_and_branches_to_diff_locked."""
        exit_stack = contextlib.ExitStack()
        self.addCleanup(exit_stack.close)
        return diff.get_trees_and_branches_to_diff_locked(
            path_list, revision_specs, old_url, new_url, exit_stack
        )

    def test_basic(self):
        tree = self.make_branch_and_tree("tree")
        (old_tree, new_tree, old_branch, new_branch, specific_files, extra_trees) = (
            self.call_gtabtd(["tree"], None, None, None)
        )

        self.assertIsInstance(old_tree, revisiontree.RevisionTree)
        self.assertEqual(_mod_revision.NULL_REVISION, old_tree.get_revision_id())
        self.assertEqual(tree.basedir, new_tree.basedir)
        self.assertEqual(tree.branch.base, old_branch.base)
        self.assertEqual(tree.branch.base, new_branch.base)
        self.assertIs(None, specific_files)
        self.assertIs(None, extra_trees)

    def test_with_rev_specs(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree_contents([("tree/file", b"oldcontent")])
        tree.add("file", ids=b"file-id")
        tree.commit("old tree", timestamp=0, rev_id=b"old-id")
        self.build_tree_contents([("tree/file", b"newcontent")])
        tree.commit("new tree", timestamp=0, rev_id=b"new-id")

        revisions = [
            revisionspec.RevisionSpec.from_string("1"),
            revisionspec.RevisionSpec.from_string("2"),
        ]
        (old_tree, new_tree, old_branch, new_branch, specific_files, extra_trees) = (
            self.call_gtabtd(["tree"], revisions, None, None)
        )

        self.assertIsInstance(old_tree, revisiontree.RevisionTree)
        self.assertEqual(b"old-id", old_tree.get_revision_id())
        self.assertIsInstance(new_tree, revisiontree.RevisionTree)
        self.assertEqual(b"new-id", new_tree.get_revision_id())
        self.assertEqual(tree.branch.base, old_branch.base)
        self.assertEqual(tree.branch.base, new_branch.base)
        self.assertIs(None, specific_files)
        self.assertEqual(tree.basedir, extra_trees[0].basedir)



# Copyright (C) 2007, 2009-2012 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os
import smtplib

from breezy import gpg, merge_directive, tests, workingtree

EMAIL1 = """From: "J. Random Hacker" <jrandom@example.com>
Subject: bar
To: pqm@example.com
User-Agent: Bazaar \\(.*\\)

# Bazaar merge directive format 2 \\(Bazaar 0.90\\)
# revision_id: bar-id
# target_branch: ../tree2
# testament_sha1: .*
# timestamp: .*
# source_branch: .
#"""


class TestMergeDirective(tests.TestCaseWithTransport):
    def prepare_merge_directive(self):
        self.tree1 = self.make_branch_and_tree("tree1")
        self.build_tree_contents([("tree1/file", b"a\nb\nc\nd\n")])
        self.tree1.branch.get_config_stack().set(
            "email", "J. Random Hacker <jrandom@example.com>"
        )
        self.tree1.add("file")
        self.tree1.commit("foo", rev_id=b"foo-id")
        self.tree2 = self.tree1.controldir.sprout("tree2").open_workingtree()
        self.build_tree_contents([("tree1/file", b"a\nb\nc\nd\ne\n")])
        self.tree1.commit("bar", rev_id=b"bar-id")
        os.chdir("tree1")
        return self.tree1, self.tree2

    def test_merge_directive(self):
        self.prepare_merge_directive()
        md_text = self.run_bzr("merge-directive ../tree2")[0]
        self.assertContainsRe(md_text, "\\+e")
        md_text = self.run_bzr("merge-directive -r -2 ../tree2")[0]
        self.assertNotContainsRe(md_text, "\\+e")
        md_text = self.run_bzr("merge-directive -r -1..-2 ../tree2")[0].encode("utf-8")
        md2 = merge_directive.MergeDirective.from_lines(md_text.splitlines(True))
        self.assertEqual(b"foo-id", md2.revision_id)
        self.assertEqual(b"bar-id", md2.base_revision_id)

    def test_submit_branch(self):
        self.prepare_merge_directive()
        self.run_bzr_error(("No submit branch",), "merge-directive", retcode=3)
        self.run_bzr("merge-directive ../tree2")

    def test_public_branch(self):
        self.prepare_merge_directive()
        self.run_bzr_error(
            ("No public branch",), "merge-directive --diff ../tree2", retcode=3
        )
        md_text = self.run_bzr("merge-directive ../tree2")[0]
        self.assertNotContainsRe(md_text, "source_branch:")
        self.run_bzr("merge-directive --diff ../tree2 .")
        self.run_bzr("merge-directive --diff")[0]
        self.assertNotContainsRe(md_text, "source_branch:")

    def test_patch_types(self):
        self.prepare_merge_directive()
        md_text = self.run_bzr("merge-directive ../tree2")[0]
        self.assertContainsRe(md_text, "# Begin bundle")
        self.assertContainsRe(md_text, "\\+e")
        md_text = self.run_bzr("merge-directive ../tree2 --diff .")[0]
        self.assertNotContainsRe(md_text, "# Begin bundle")
        self.assertContainsRe(md_text, "\\+e")
        md_text = self.run_bzr("merge-directive --plain")[0]
        self.assertNotContainsRe(md_text, "\\+e")

    def test_message(self):
        self.prepare_merge_directive()
        md_text = self.run_bzr("merge-directive ../tree2")[0]
        self.assertNotContainsRe(md_text, "message: Message for merge")
        md_text = self.run_bzr("merge-directive -m Message_for_merge")[0]
        self.assertContainsRe(md_text, "message: Message_for_merge")

    def test_signing(self):
        self.prepare_merge_directive()
        old_strategy = gpg.GPGStrategy
        gpg.GPGStrategy = gpg.LoopbackGPGStrategy
        try:
            md_text = self.run_bzr("merge-directive --sign ../tree2")[0]
        finally:
            gpg.GPGStrategy = old_strategy
        self.assertContainsRe(md_text, "^-----BEGIN PSEUDO-SIGNED CONTENT")

    def run_bzr_fakemail(self, *args, **kwargs):
        sendmail_calls = []

        def sendmail(self, from_, to, message):
            sendmail_calls.append((self, from_, to, message))

        connect_calls = []

        def connect(self, host="localhost", port=0):
            connect_calls.append((self, host, port))
            return (220, "Ok")

        def has_extn(self, extension):
            return False

        def ehlo(self):
            return (200, "Ok")

        old_sendmail = smtplib.SMTP.sendmail
        smtplib.SMTP.sendmail = sendmail
        old_connect = smtplib.SMTP.connect
        smtplib.SMTP.connect = connect
        old_ehlo = smtplib.SMTP.ehlo
        smtplib.SMTP.ehlo = ehlo
        old_has_extn = smtplib.SMTP.has_extn
        smtplib.SMTP.has_extn = has_extn
        try:
            result = self.run_bzr(*args, **kwargs)
        finally:
            smtplib.SMTP.sendmail = old_sendmail
            smtplib.SMTP.connect = old_connect
            smtplib.SMTP.ehlo = old_ehlo
            smtplib.SMTP.has_extn = old_has_extn
        return result + (connect_calls, sendmail_calls)

    def test_mail_default(self):
        tree1, tree2 = self.prepare_merge_directive()
        md_text, errr, connect_calls, sendmail_calls = self.run_bzr_fakemail(
            [
                "merge-directive",
                "--mail-to",
                "pqm@example.com",
                "--plain",
                "../tree2",
                ".",
            ]
        )
        self.assertEqual("", md_text)
        self.assertEqual(1, len(connect_calls))
        call = connect_calls[0]
        self.assertEqual(("localhost", 0), call[1:3])
        self.assertEqual(1, len(sendmail_calls))
        call = sendmail_calls[0]
        self.assertEqual(("jrandom@example.com", ["pqm@example.com"]), call[1:3])
        self.assertContainsRe(call[3], EMAIL1)

    def test_pull_raw(self):
        self.prepare_merge_directive()
        self.tree1.commit("baz", rev_id=b"baz-id")
        md_text = self.run_bzr(
            [
                "merge-directive",
                self.tree2.basedir,
                "-r",
                "2",
                self.tree1.basedir,
                "--plain",
            ]
        )[0]
        self.build_tree_contents([("../directive", md_text)])
        os.chdir("../tree2")
        self.run_bzr("pull ../directive")
        wt = workingtree.WorkingTree.open(".")
        self.assertEqual(b"bar-id", wt.last_revision())

    def test_pull_user_r(self):
        """If the user supplies -r, an error is emitted."""
        self.prepare_merge_directive()
        self.tree1.commit("baz", rev_id=b"baz-id")
        md_text = self.run_bzr(
            ["merge-directive", self.tree2.basedir, self.tree1.basedir, "--plain"]
        )[0]
        self.build_tree_contents([("../directive", md_text)])
        os.chdir("../tree2")
        self.run_bzr_error(
            ("Cannot use -r with merge directives or bundles",),
            "pull -r 2 ../directive",
        )

    def test_pull_bundle(self):
        self.prepare_merge_directive()
        self.tree1.commit("baz", rev_id=b"baz-id")
        md_text = self.run_bzr(
            ["merge-directive", self.tree2.basedir, "-r", "2", "/dev/null", "--bundle"]
        )[0]
        self.build_tree_contents([("../directive", md_text)])
        os.chdir("../tree2")
        self.run_bzr("pull ../directive")
        wt = workingtree.WorkingTree.open(".")
        self.assertEqual(b"bar-id", wt.last_revision())

    def test_merge_raw(self):
        self.prepare_merge_directive()
        self.tree1.commit("baz", rev_id=b"baz-id")
        md_text = self.run_bzr(
            [
                "merge-directive",
                self.tree2.basedir,
                "-r",
                "2",
                self.tree1.basedir,
                "--plain",
            ]
        )[0]
        self.build_tree_contents([("../directive", md_text)])
        os.chdir("../tree2")
        self.run_bzr("merge ../directive")
        wt = workingtree.WorkingTree.open(".")
        self.assertEqual(b"bar-id", wt.get_parent_ids()[1])

    def test_merge_user_r(self):
        """If the user supplies -r, an error is emitted."""
        self.prepare_merge_directive()
        self.tree1.commit("baz", rev_id=b"baz-id")
        md_text = self.run_bzr(
            ["merge-directive", self.tree2.basedir, self.tree1.basedir, "--plain"]
        )[0]
        self.build_tree_contents([("../directive", md_text)])
        os.chdir("../tree2")
        self.run_bzr_error(
            ("Cannot use -r with merge directives or bundles",),
            "merge -r 2 ../directive",
        )

    def test_merge_bundle(self):
        self.prepare_merge_directive()
        self.tree1.commit("baz", rev_id=b"baz-id")
        md_text = self.run_bzr(
            ["merge-directive", self.tree2.basedir, "-r", "2", "/dev/null", "--bundle"]
        )[0]
        self.build_tree_contents([("../directive", md_text)])
        os.chdir("../tree2")
        self.run_bzr("merge ../directive")
        wt = workingtree.WorkingTree.open(".")
        self.assertEqual(b"bar-id", wt.get_parent_ids()[1])

    def test_mail_uses_config(self):
        tree1, tree2 = self.prepare_merge_directive()
        br = tree1.branch
        br.get_config_stack().set("smtp_server", "bogushost")
        md_text, errr, connect_calls, sendmail_calls = self.run_bzr_fakemail(
            "merge-directive --mail-to pqm@example.com --plain ../tree2 ."
        )
        call = connect_calls[0]
        self.assertEqual(("bogushost", 0), call[1:3])

    def test_no_common_ancestor(self):
        foo = self.make_branch_and_tree("foo")
        foo.commit("rev1")
        self.make_branch_and_tree("bar")
        self.run_bzr("merge-directive ../bar", working_dir="foo")

    def test_no_commits(self):
        self.make_branch_and_tree("foo")
        self.make_branch_and_tree("bar")
        self.run_bzr_error(
            ("No revisions to bundle.",), "merge-directive ../bar", working_dir="foo"
        )

    def test_encoding_exact(self):
        tree1, tree2 = self.prepare_merge_directive()
        tree1.commit("messag\xe9")
        self.run_bzr("merge-directive ../tree2")  # no exception raised

    def test_merge_directive_directory(self):
        """Test --directory option."""
        import re

        re_timestamp = re.compile(r"^# timestamp: .*", re.M)
        self.prepare_merge_directive()
        md1 = self.run_bzr("merge-directive ../tree2")[0]
        md1 = re_timestamp.sub("# timestamp: XXX", md1)
        os.chdir("..")
        md2 = self.run_bzr("merge-directive --directory tree1 tree2")[0]
        md2 = re_timestamp.sub("# timestamp: XXX", md2)
        self.assertEqualDiff(md1.replace("../tree2", "tree2"), md2)



# Copyright (C) 2005-2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

# Mr. Smoketoomuch: I'm sorry?
# Mr. Bounder: You'd better cut down a little then.
# Mr. Smoketoomuch: Oh, I see! Smoke too much so I'd better cut down a little
#                   then!

"""Black-box tests for bzr.

These check that it behaves properly when it's invoked through the regular
command-line interface. This doesn't actually run a new interpreter but
rather starts again from the run_brz function.
"""


# XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
# Note: Please don't add new tests here, it's too big and bulky.  Instead add
# them into small suites in breezy.tests.blackbox.test_FOO for the particular
# UI command/aspect that is being tested.

import os
import sys

from breezy import osutils
from breezy.branch import Branch
from breezy.tests import TestCaseWithTransport
from breezy.tests.http_utils import TestCaseWithWebserver
from breezy.tests.test_sftp_transport import TestCaseWithSFTPServer


class TestCommands(TestCaseWithTransport):
    def test_invalid_commands(self):
        self.run_bzr("pants", retcode=3)
        self.run_bzr("--pants off", retcode=3)
        self.run_bzr("diff --message foo", retcode=3)

    def test_revert(self):
        self.run_bzr("init")

        with open("hello", "w") as f:
            f.write("foo")
        self.run_bzr("add hello")
        self.run_bzr("commit -m setup hello")

        with open("goodbye", "w") as f:
            f.write("baz")
        self.run_bzr("add goodbye")
        self.run_bzr("commit -m setup goodbye")

        with open("hello", "w") as f:
            f.write("bar")
        with open("goodbye", "w") as f:
            f.write("qux")
        self.run_bzr("revert hello")
        self.check_file_contents("hello", b"foo")
        self.check_file_contents("goodbye", b"qux")
        self.run_bzr("revert")
        self.check_file_contents("goodbye", b"baz")

        os.mkdir("revertdir")
        self.run_bzr("add revertdir")
        self.run_bzr("commit -m f")
        os.rmdir("revertdir")
        self.run_bzr("revert")

        if osutils.supports_symlinks(self.test_dir):
            os.symlink("/unlikely/to/exist", "symlink")
            self.run_bzr("add symlink")
            self.run_bzr("commit -m f")
            os.unlink("symlink")
            self.run_bzr("revert")
            self.assertPathExists("symlink")
            os.unlink("symlink")
            os.symlink("a-different-path", "symlink")
            self.run_bzr("revert")
            self.assertEqual("/unlikely/to/exist", os.readlink("symlink"))
        else:
            self.log("skipping revert symlink tests")

        with open("hello", "w") as f:
            f.write("xyz")
        self.run_bzr("commit -m xyz hello")
        self.run_bzr("revert -r 1 hello")
        self.check_file_contents("hello", b"foo")
        self.run_bzr("revert hello")
        self.check_file_contents("hello", b"xyz")
        os.chdir("revertdir")
        self.run_bzr("revert")
        os.chdir("..")

    def example_branch(self):
        self.run_bzr("init")
        with open("hello", "w") as f:
            f.write("foo")
        self.run_bzr("add hello")
        self.run_bzr("commit -m setup hello")
        with open("goodbye", "w") as f:
            f.write("baz")
        self.run_bzr("add goodbye")
        self.run_bzr("commit -m setup goodbye")

    def test_pull_verbose(self):
        """Pull changes from one branch to another and watch the output."""
        os.mkdir("a")
        os.chdir("a")

        self.example_branch()

        os.chdir("..")
        self.run_bzr("branch a b")
        os.chdir("b")
        with open("b", "wb") as f:
            f.write(b"else\n")
        self.run_bzr("add b")
        self.run_bzr(["commit", "-m", "added b"])

        os.chdir("../a")
        out = self.run_bzr("pull --verbose ../b")[0]
        self.assertNotEqual(out.find("Added Revisions:"), -1)
        self.assertNotEqual(out.find("message:\n  added b"), -1)
        self.assertNotEqual(out.find("added b"), -1)

        # Check that --overwrite --verbose prints out the removed entries
        self.run_bzr("commit -m foo --unchanged")
        os.chdir("../b")
        self.run_bzr("commit -m baz --unchanged")
        self.run_bzr("pull ../a", retcode=3)
        out = self.run_bzr("pull --overwrite --verbose ../a")[0]

        remove_loc = out.find("Removed Revisions:")
        self.assertNotEqual(remove_loc, -1)
        added_loc = out.find("Added Revisions:")
        self.assertNotEqual(added_loc, -1)

        removed_message = out.find("message:\n  baz")
        self.assertNotEqual(removed_message, -1)
        self.assertTrue(remove_loc < removed_message < added_loc)

        added_message = out.find("message:\n  foo")
        self.assertNotEqual(added_message, -1)
        self.assertTrue(added_loc < added_message)

    def test_locations(self):
        """Using and remembering different locations."""
        os.mkdir("a")
        os.chdir("a")
        self.run_bzr("init")
        self.run_bzr("commit -m unchanged --unchanged")
        self.run_bzr("pull", retcode=3)
        self.run_bzr("merge", retcode=3)
        self.run_bzr("branch . ../b")
        os.chdir("../b")
        self.run_bzr("pull")
        self.run_bzr("branch . ../c")
        self.run_bzr("pull ../c")
        self.run_bzr("merge")
        os.chdir("../a")
        self.run_bzr("pull ../b")
        self.run_bzr("pull")
        self.run_bzr("pull ../c")
        self.run_bzr("branch ../c ../d")
        osutils.rmtree("../c")
        self.run_bzr("pull")
        os.chdir("../b")
        self.run_bzr("pull")
        os.chdir("../d")
        self.run_bzr("pull", retcode=3)
        self.run_bzr("pull ../a --remember")
        self.run_bzr("pull")

    def test_unknown_command(self):
        """Handling of unknown command."""
        out, err = self.run_bzr("fluffy-badger", retcode=3)
        self.assertEqual(out, "")
        err.index("unknown command")

    def create_conflicts(self):
        """Create a conflicted tree."""
        os.mkdir("base")
        os.chdir("base")
        with open("hello", "wb") as f:
            f.write(b"hi world")
        with open("answer", "wb") as f:
            f.write(b"42")
        self.run_bzr("init")
        self.run_bzr("add")
        self.run_bzr("commit -m base")
        self.run_bzr("branch . ../other")
        self.run_bzr("branch . ../this")
        os.chdir("../other")
        with open("hello", "wb") as f:
            f.write(b"Hello.")
        with open("answer", "wb") as f:
            f.write(b"Is anyone there?")
        self.run_bzr("commit -m other")
        os.chdir("../this")
        with open("hello", "wb") as f:
            f.write(b"Hello, world")
        self.run_bzr("mv answer question")
        with open("question", "wb") as f:
            f.write(b"What do you get when you multiply sixtimes nine?")
        self.run_bzr("commit -m this")

    def test_status(self):
        os.mkdir("branch1")
        os.chdir("branch1")
        self.run_bzr("init")
        self.run_bzr("commit --unchanged --message f")
        self.run_bzr("branch . ../branch2")
        self.run_bzr("branch . ../branch3")
        self.run_bzr("commit --unchanged --message peter")
        os.chdir("../branch2")
        self.run_bzr("merge ../branch1")
        self.run_bzr("commit --unchanged --message pumpkin")
        os.chdir("../branch3")
        self.run_bzr("merge ../branch2")
        self.run_bzr("status")[0]

    def test_conflicts(self):
        """Handling of merge conflicts."""
        self.create_conflicts()
        self.run_bzr("merge ../other --show-base", retcode=1)
        with open("hello") as f:
            conflict_text = f.read()
        self.assertTrue("<<<<<<<" in conflict_text)
        self.assertTrue(">>>>>>>" in conflict_text)
        self.assertTrue("=======" in conflict_text)
        self.assertTrue("|||||||" in conflict_text)
        self.assertTrue("hi world" in conflict_text)
        self.run_bzr("revert")
        self.run_bzr("resolve --all")
        self.run_bzr("merge ../other", retcode=1)
        with open("hello") as f:
            conflict_text = f.read()
        self.assertTrue("|||||||" not in conflict_text)
        self.assertTrue("hi world" not in conflict_text)
        result = self.run_bzr("conflicts")[0]
        self.assertEqual(result, "Text conflict in hello\nText conflict in question\n")
        result = self.run_bzr("status")[0]
        self.assertTrue(
            "conflicts:\n  Text conflict in hello\n"
            "  Text conflict in question\n" in result,
            result,
        )
        self.run_bzr("resolve hello")
        result = self.run_bzr("conflicts")[0]
        self.assertEqual(result, "Text conflict in question\n")
        self.run_bzr("commit -m conflicts", retcode=3)
        self.run_bzr("resolve --all")
        result = self.run_bzr("conflicts")[0]
        self.run_bzr("commit -m conflicts")
        self.assertEqual(result, "")

    def test_push(self):
        # create a source branch
        os.mkdir("my-branch")
        os.chdir("my-branch")
        self.example_branch()

        # with no push target, fail
        self.run_bzr("push", retcode=3)
        # with an explicit target work
        self.run_bzr("push ../output-branch")
        # with an implicit target work
        self.run_bzr("push")
        # nothing missing
        self.run_bzr("missing ../output-branch")
        # advance this branch
        self.run_bzr("commit --unchanged -m unchanged")

        os.chdir("../output-branch")
        # There is no longer a difference as long as we have
        # access to the working tree
        self.run_bzr("diff")

        # But we should be missing a revision
        self.run_bzr("missing ../my-branch", retcode=1)

        # diverge the branches
        self.run_bzr("commit --unchanged -m unchanged")
        os.chdir("../my-branch")
        # cannot push now
        self.run_bzr("push", retcode=3)
        # and there are difference
        self.run_bzr("missing ../output-branch", retcode=1)
        self.run_bzr("missing --verbose ../output-branch", retcode=1)
        # but we can force a push
        self.run_bzr("push --overwrite")
        # nothing missing
        self.run_bzr("missing ../output-branch")

        # pushing to a new dir with no parent should fail
        self.run_bzr("push ../missing/new-branch", retcode=3)
        # unless we provide --create-prefix
        self.run_bzr("push --create-prefix ../missing/new-branch")
        # nothing missing
        self.run_bzr("missing ../missing/new-branch")

    def test_external_command(self):
        """Test that external commands can be run by setting the path."""
        # We don't at present run brz in a subprocess for blackbox tests, and so
        # don't really capture stdout, only the internal python stream.
        # Therefore we don't use a subcommand that produces any output or does
        # anything -- we just check that it can be run successfully.
        cmd_name = "test-command"
        if sys.platform == "win32":
            cmd_name += ".bat"
        self.overrideEnv("BZRPATH", None)

        f = open(cmd_name, "wb")
        if sys.platform == "win32":
            f.write(b"@echo off\n")
        else:
            f.write(b"#!/bin/sh\n")
        # f.write('echo Hello from test-command')
        f.close()
        os.chmod(cmd_name, 0o755)

        # It should not find the command in the local
        # directory by default, since it is not in my path
        self.run_bzr(cmd_name, retcode=3)

        # Now put it into my path
        self.overrideEnv("BZRPATH", ".")
        self.run_bzr(cmd_name)

        # Make sure empty path elements are ignored
        self.overrideEnv("BZRPATH", os.pathsep)
        self.run_bzr(cmd_name, retcode=3)


def listdir_sorted(dir):
    L = sorted(os.listdir(dir))
    return L


class OldTests(TestCaseWithTransport):
    """old tests moved from ./testbzr."""

    def test_bzr(self):
        from os import chdir, mkdir

        progress = self.log

        progress("basic branch creation")
        mkdir("branch1")
        chdir("branch1")
        self.run_bzr("init")

        self.assertIsSameRealPath(
            self.run_bzr("root")[0].rstrip(), osutils.pathjoin(self.test_dir, "branch1")
        )

        progress("status of new file")

        with open("test.txt", "w") as f:
            f.write("hello world!\n")

        self.assertEqual(self.run_bzr("unknowns")[0], "test.txt\n")

        out = self.run_bzr("status")[0]
        self.assertEqual(out, "unknown:\n  test.txt\n")

        with open("test2.txt", "w") as f:
            f.write("goodbye cruel world...\n")

        out = self.run_bzr("status test.txt")[0]
        self.assertEqual(out, "unknown:\n  test.txt\n")

        out = self.run_bzr("status")[0]
        self.assertEqual(out, ("unknown:\n  test.txt\n  test2.txt\n"))

        os.unlink("test2.txt")

        progress("command aliases")
        out = self.run_bzr("st")[0]
        self.assertEqual(out, ("unknown:\n  test.txt\n"))

        out = self.run_bzr("stat")[0]
        self.assertEqual(out, ("unknown:\n  test.txt\n"))

        progress("command help")
        self.run_bzr("help st")
        self.run_bzr("help")
        self.run_bzr("help commands")
        self.run_bzr("help slartibartfast", retcode=3)

        out = self.run_bzr("help ci")[0]
        out.index("Aliases:  ci, checkin\n")

        with open("hello.txt", "w") as f:
            f.write("some nice new content\n")

        self.run_bzr("add hello.txt")

        with open("msg.tmp", "w") as f:
            f.write("this is my new commit\nand it has multiple lines, for fun")

        self.run_bzr("commit -F msg.tmp")

        self.assertEqual(self.run_bzr("revno")[0], "1\n")
        self.run_bzr("export -r 1 export-1.tmp")
        self.run_bzr("export export.tmp")

        self.run_bzr("log")
        self.run_bzr("log -v")
        self.run_bzr("log -v --forward")
        self.run_bzr("log -m", retcode=3)
        log_out = self.run_bzr("log -m commit")[0]
        self.assertTrue("this is my new commit\n  and" in log_out)
        self.assertTrue("rename nested" not in log_out)
        self.assertTrue("revision-id" not in log_out)
        self.assertTrue("revision-id" in self.run_bzr("log --show-ids -m commit")[0])

        log_out = self.run_bzr("log --line")[0]
        # determine the widest line we want
        max_width = osutils.terminal_width()
        if max_width is not None:
            for line in log_out.splitlines():
                self.assertTrue(len(line) <= max_width - 1, len(line))
        self.assertTrue("this is my new commit and" not in log_out)
        self.assertTrue("this is my new commit" in log_out)

        progress("file with spaces in name")
        mkdir("sub directory")
        with open("sub directory/file with spaces ", "w") as f:
            f.write("see how this works\n")
        self.run_bzr("add .")
        self.run_bzr("diff", retcode=1)
        self.run_bzr("commit -m add-spaces")
        self.run_bzr("check")

        self.run_bzr("log")
        self.run_bzr("log --forward")

        self.run_bzr("info")

        if osutils.supports_symlinks(self.test_dir):
            progress("symlinks")
            mkdir("symlinks")
            chdir("symlinks")
            self.run_bzr("init")
            os.symlink("NOWHERE1", "link1")
            self.run_bzr("add link1")
            self.assertEqual(self.run_bzr("unknowns")[0], "")
            self.run_bzr(["commit", "-m", "1: added symlink link1"])

            mkdir("d1")
            self.run_bzr("add d1")
            self.assertEqual(self.run_bzr("unknowns")[0], "")
            os.symlink("NOWHERE2", "d1/link2")
            self.assertEqual(self.run_bzr("unknowns")[0], "d1/link2\n")
            # is d1/link2 found when adding d1
            self.run_bzr("add d1")
            self.assertEqual(self.run_bzr("unknowns")[0], "")
            os.symlink("NOWHERE3", "d1/link3")
            self.assertEqual(self.run_bzr("unknowns")[0], "d1/link3\n")
            self.run_bzr(["commit", "-m", "2: added dir, symlink"])

            self.run_bzr("rename d1 d2")
            self.run_bzr("move d2/link2 .")
            self.run_bzr("move link1 d2")
            self.assertEqual(os.readlink("./link2"), "NOWHERE2")
            self.assertEqual(os.readlink("d2/link1"), "NOWHERE1")
            self.run_bzr("add d2/link3")
            self.run_bzr("diff", retcode=1)
            self.run_bzr(["commit", "-m", "3: rename of dir, move symlinks, add link3"])

            os.unlink("link2")
            os.symlink("TARGET 2", "link2")
            os.unlink("d2/link1")
            os.symlink("TARGET 1", "d2/link1")
            self.run_bzr("diff", retcode=1)
            self.assertEqual(self.run_bzr("relpath d2/link1")[0], "d2/link1\n")
            self.run_bzr(["commit", "-m", "4: retarget of two links"])

            self.run_bzr("remove --keep d2/link1")
            self.assertEqual(self.run_bzr("unknowns")[0], "d2/link1\n")
            self.run_bzr(["commit", "-m", "5: remove d2/link1"])
            # try with the rm alias
            self.run_bzr("add d2/link1")
            self.run_bzr(["commit", "-m", "6: add d2/link1"])
            self.run_bzr("rm --keep d2/link1")
            self.assertEqual(self.run_bzr("unknowns")[0], "d2/link1\n")
            self.run_bzr(["commit", "-m", "7: remove d2/link1"])

            os.mkdir("d1")
            self.run_bzr("add d1")
            self.run_bzr("rename d2/link3 d1/link3new")
            self.assertEqual(self.run_bzr("unknowns")[0], "d2/link1\n")
            self.run_bzr(["commit", "-m", "8: remove d2/link1, move/rename link3"])

            self.run_bzr("check")

            self.run_bzr("export -r 1 exp1.tmp")
            chdir("exp1.tmp")
            self.assertEqual(listdir_sorted("."), ["link1"])
            self.assertEqual(os.readlink("link1"), "NOWHERE1")
            chdir("..")

            self.run_bzr("export -r 2 exp2.tmp")
            chdir("exp2.tmp")
            self.assertEqual(listdir_sorted("."), ["d1", "link1"])
            chdir("..")

            self.run_bzr("export -r 3 exp3.tmp")
            chdir("exp3.tmp")
            self.assertEqual(listdir_sorted("."), ["d2", "link2"])
            self.assertEqual(listdir_sorted("d2"), ["link1", "link3"])
            self.assertEqual(os.readlink("d2/link1"), "NOWHERE1")
            self.assertEqual(os.readlink("link2"), "NOWHERE2")
            chdir("..")

            self.run_bzr("export -r 4 exp4.tmp")
            chdir("exp4.tmp")
            self.assertEqual(listdir_sorted("."), ["d2", "link2"])
            self.assertEqual(os.readlink("d2/link1"), "TARGET 1")
            self.assertEqual(os.readlink("link2"), "TARGET 2")
            self.assertEqual(listdir_sorted("d2"), ["link1", "link3"])
            chdir("..")

            self.run_bzr("export -r 5 exp5.tmp")
            chdir("exp5.tmp")
            self.assertEqual(listdir_sorted("."), ["d2", "link2"])
            self.assertTrue(os.path.islink("link2"))
            self.assertTrue(listdir_sorted("d2") == ["link3"])
            chdir("..")

            self.run_bzr("export -r 8 exp6.tmp")
            chdir("exp6.tmp")
            self.assertEqual(listdir_sorted("."), ["d1", "d2", "link2"])
            self.assertEqual(listdir_sorted("d1"), ["link3new"])
            self.assertEqual(listdir_sorted("d2"), [])
            self.assertEqual(os.readlink("d1/link3new"), "NOWHERE3")
            chdir("..")
        else:
            progress("skipping symlink tests")


class RemoteTests:
    """Test brz ui commands against remote branches."""

    def test_branch(self):
        os.mkdir("from")
        wt = self.make_branch_and_tree("from")
        branch = wt.branch
        wt.commit("empty commit for nonsense", allow_pointless=True)
        url = self.get_readonly_url("from")
        self.run_bzr(["branch", url, "to"])
        branch = Branch.open("to")
        self.assertEqual(1, branch.last_revision_info()[0])
        # the branch should be set in to to from
        self.assertEqual(url + "/", branch.get_parent())

    def test_log(self):
        self.build_tree(["branch/", "branch/file"])
        self.run_bzr("init branch")[0]
        self.run_bzr("add branch/file")[0]
        self.run_bzr("commit -m foo branch")[0]
        url = self.get_readonly_url("branch/file")
        output = self.run_bzr("log {}".format(url))[0]
        self.assertEqual(8, len(output.split("\n")))

    def test_check(self):
        self.build_tree(["branch/", "branch/file"])
        self.run_bzr("init branch")[0]
        self.run_bzr("add branch/file")[0]
        self.run_bzr("commit -m foo branch")[0]
        url = self.get_readonly_url("branch/")
        self.run_bzr(["check", url])

    def test_push(self):
        # create a source branch
        os.mkdir("my-branch")
        os.chdir("my-branch")
        self.run_bzr("init")
        with open("hello", "w") as f:
            f.write("foo")
        self.run_bzr("add hello")
        self.run_bzr("commit -m setup")

        # with an explicit target work
        self.run_bzr(["push", self.get_url("output-branch")])


class HTTPTests(TestCaseWithWebserver, RemoteTests):
    """Test various commands against a HTTP server."""


class SFTPTestsAbsolute(TestCaseWithSFTPServer, RemoteTests):
    """Test various commands against a SFTP server using abs paths."""


class SFTPTestsAbsoluteSibling(TestCaseWithSFTPServer, RemoteTests):
    """Test various commands against a SFTP server using abs paths."""

    def setUp(self):
        super().setUp()
        self._override_home = "/dev/noone/runs/tests/here"


class SFTPTestsRelative(TestCaseWithSFTPServer, RemoteTests):
    """Test various commands against a SFTP server using homedir rel paths."""

    def setUp(self):
        super().setUp()
        self._get_remote_is_absolute = False



# Copyright (C) 2006-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Black-box tests for brz log."""

import os

from breezy import branchbuilder, errors, log, osutils, tests
from breezy.tests import features, test_log


class TestLog(tests.TestCaseWithTransport, test_log.TestLogMixin):
    def make_minimal_branch(self, path=".", format=None):
        tree = self.make_branch_and_tree(path, format=format)
        self.build_tree([path + "/hello.txt"])
        tree.add("hello.txt")
        tree.commit(message="message1")
        return tree

    def make_linear_branch(self, path=".", format=None):
        tree = self.make_branch_and_tree(path, format=format)
        self.build_tree(
            [path + "/hello.txt", path + "/goodbye.txt", path + "/meep.txt"]
        )
        tree.add("hello.txt")
        tree.commit(message="message1")
        tree.add("goodbye.txt")
        tree.commit(message="message2")
        tree.add("meep.txt")
        tree.commit(message="message3")
        return tree

    def make_merged_branch(self, path=".", format=None):
        tree = self.make_linear_branch(path, format)
        tree2 = tree.controldir.sprout(
            "tree2", revision_id=tree.branch.get_rev_id(1)
        ).open_workingtree()
        tree2.commit(message="tree2 message2")
        tree2.commit(message="tree2 message3")
        tree.merge_from_branch(tree2.branch)
        tree.commit(message="merge")
        return tree


class TestLogWithLogCatcher(TestLog):
    def setUp(self):
        super().setUp()
        # Capture log formatter creations

        class MyLogFormatter(test_log.LogCatcher):
            def __new__(klass, *args, **kwargs):
                self.log_catcher = test_log.LogCatcher(*args, **kwargs)
                # Always return our own log formatter
                return self.log_catcher

        # Break cycle with closure over self on cleanup by removing method
        self.addCleanup(setattr, MyLogFormatter, "__new__", None)

        def getme(branch):
            # Always return our own log formatter class hijacking the
            # default behavior (which requires setting up a config
            # variable)
            return MyLogFormatter

        self.overrideAttr(log.log_formatter_registry, "get_default", getme)

    def get_captured_revisions(self):
        return self.log_catcher.revisions

    def assertLogRevnos(self, args, expected_revnos, working_dir=".", out="", err=""):
        actual_out, actual_err = self.run_bzr(["log"] + args, working_dir=working_dir)
        self.assertEqual(out, actual_out)
        self.assertEqual(err, actual_err)
        self.assertEqual(
            expected_revnos, [r.revno for r in self.get_captured_revisions()]
        )

    def assertLogRevnosAndDepths(
        self, args, expected_revnos_and_depths, working_dir="."
    ):
        self.run_bzr(["log"] + args, working_dir=working_dir)
        self.assertEqual(
            expected_revnos_and_depths,
            [(r.revno, r.merge_depth) for r in self.get_captured_revisions()],
        )


class TestLogRevSpecs(TestLogWithLogCatcher):
    def test_log_no_revspec(self):
        self.make_linear_branch()
        self.assertLogRevnos([], ["3", "2", "1"])

    def test_log_null_end_revspec(self):
        self.make_linear_branch()
        self.assertLogRevnos(["-r1.."], ["3", "2", "1"])

    def test_log_null_begin_revspec(self):
        self.make_linear_branch()
        self.assertLogRevnos(["-r..3"], ["3", "2", "1"])

    def test_log_null_both_revspecs(self):
        self.make_linear_branch()
        self.assertLogRevnos(["-r.."], ["3", "2", "1"])

    def test_log_negative_begin_revspec_full_log(self):
        self.make_linear_branch()
        self.assertLogRevnos(["-r-3.."], ["3", "2", "1"])

    def test_log_negative_both_revspec_full_log(self):
        self.make_linear_branch()
        self.assertLogRevnos(["-r-3..-1"], ["3", "2", "1"])

    def test_log_negative_both_revspec_partial(self):
        self.make_linear_branch()
        self.assertLogRevnos(["-r-3..-2"], ["2", "1"])

    def test_log_negative_begin_revspec(self):
        self.make_linear_branch()
        self.assertLogRevnos(["-r-2.."], ["3", "2"])

    def test_log_positive_revspecs(self):
        self.make_linear_branch()
        self.assertLogRevnos(["-r1..3"], ["3", "2", "1"])

    def test_log_dotted_revspecs(self):
        self.make_merged_branch()
        self.assertLogRevnos(["-n0", "-r1..1.1.1"], ["1.1.1", "1"])

    def test_log_limit(self):
        tree = self.make_branch_and_tree(".")
        # We want more commits than our batch size starts at
        for pos in range(10):
            tree.commit("{}".format(pos))
        self.assertLogRevnos(["--limit", "2"], ["10", "9"])

    def test_log_limit_short(self):
        self.make_linear_branch()
        self.assertLogRevnos(["-l", "2"], ["3", "2"])

    def test_log_change_revno(self):
        self.make_linear_branch()
        self.assertLogRevnos(["-c1"], ["1"])

    def test_branch_revspec(self):
        foo = self.make_branch_and_tree("foo")
        bar = self.make_branch_and_tree("bar")
        self.build_tree(["foo/foo.txt", "bar/bar.txt"])
        foo.add("foo.txt")
        bar.add("bar.txt")
        foo.commit(message="foo")
        bar.commit(message="bar")
        self.run_bzr("log -r branch:../bar", working_dir="foo")
        self.assertEqual(
            [bar.branch.get_rev_id(1)],
            [r.rev.revision_id for r in self.get_captured_revisions()],
        )


class TestLogExcludeCommonAncestry(TestLogWithLogCatcher):
    def test_exclude_common_ancestry_simple_revnos(self):
        self.make_linear_branch()
        self.assertLogRevnos(["-r1..3", "--exclude-common-ancestry"], ["3", "2"])


class TestLogMergedLinearAncestry(TestLogWithLogCatcher):
    def setUp(self):
        super().setUp()
        # FIXME: Using a MemoryTree would be even better here (but until we
        # stop calling run_bzr, there is no point) --vila 100118.
        builder = branchbuilder.BranchBuilder(self.get_transport())
        builder.start_series()
        # 1
        # | \
        # 2  1.1.1
        # | / |
        # 3  1.1.2
        # |   |
        # |  1.1.3
        # | / |
        # 4  1.1.4
        # | /
        # 5
        # | \
        # | 5.1.1
        # | /
        # 6

        # mainline
        builder.build_snapshot(
            None, [("add", ("", b"root-id", "directory", ""))], revision_id=b"1"
        )
        builder.build_snapshot([b"1"], [], revision_id=b"2")
        # branch
        builder.build_snapshot([b"1"], [], revision_id=b"1.1.1")
        # merge branch into mainline
        builder.build_snapshot([b"2", b"1.1.1"], [], revision_id=b"3")
        # new commits in branch
        builder.build_snapshot([b"1.1.1"], [], revision_id=b"1.1.2")
        builder.build_snapshot([b"1.1.2"], [], revision_id=b"1.1.3")
        # merge branch into mainline
        builder.build_snapshot([b"3", b"1.1.3"], [], revision_id=b"4")
        # merge mainline into branch
        builder.build_snapshot([b"1.1.3", b"4"], [], revision_id=b"1.1.4")
        # merge branch into mainline
        builder.build_snapshot([b"4", b"1.1.4"], [], revision_id=b"5")
        builder.build_snapshot([b"5"], [], revision_id=b"5.1.1")
        builder.build_snapshot([b"5", b"5.1.1"], [], revision_id=b"6")
        builder.finish_series()

    def test_n0(self):
        self.assertLogRevnos(
            ["-n0", "-r1.1.1..1.1.4"], ["1.1.4", "4", "1.1.3", "1.1.2", "3", "1.1.1"]
        )

    def test_n0_forward(self):
        self.assertLogRevnos(
            ["-n0", "-r1.1.1..1.1.4", "--forward"],
            ["3", "1.1.1", "4", "1.1.2", "1.1.3", "1.1.4"],
        )

    def test_n1(self):
        # starting from 1.1.4 we follow the left-hand ancestry
        self.assertLogRevnos(
            ["-n1", "-r1.1.1..1.1.4"], ["1.1.4", "1.1.3", "1.1.2", "1.1.1"]
        )

    def test_n1_forward(self):
        self.assertLogRevnos(
            ["-n1", "-r1.1.1..1.1.4", "--forward"], ["1.1.1", "1.1.2", "1.1.3", "1.1.4"]
        )

    def test_fallback_when_end_rev_is_not_on_mainline(self):
        self.assertLogRevnos(
            ["-n1", "-r1.1.1..5.1.1"],
            # We don't get 1.1.1 because we say -n1
            ["5.1.1", "5", "4", "3"],
        )


class Test_GenerateAllRevisions(TestLogWithLogCatcher):
    def setUp(self):
        super().setUp()
        builder = self.make_branch_with_many_merges()
        b = builder.get_branch()
        b.lock_read()
        self.addCleanup(b.unlock)
        self.branch = b

    def make_branch_with_many_merges(self, path=".", format=None):
        builder = branchbuilder.BranchBuilder(self.get_transport())
        builder.start_series()
        # The graph below may look a bit complicated (and it may be but I've
        # banged my head enough on it) but the bug requires at least dotted
        # revnos *and* merged revisions below that.
        # 1
        # | \
        # 2  1.1.1
        # | X
        # 3  2.1.1
        # |   |    \
        # |  2.1.2  2.2.1
        # |   |    X
        # |  2.1.3  \
        # | /       /
        # 4        /
        # |       /
        # 5 -----/
        builder.build_snapshot(
            None, [("add", ("", b"root-id", "directory", ""))], revision_id=b"1"
        )
        builder.build_snapshot([b"1"], [], revision_id=b"2")
        builder.build_snapshot([b"1"], [], revision_id=b"1.1.1")
        builder.build_snapshot([b"2"], [], revision_id=b"2.1.1")
        builder.build_snapshot([b"2", b"1.1.1"], [], revision_id=b"3")
        builder.build_snapshot([b"2.1.1"], [], revision_id=b"2.1.2")
        builder.build_snapshot([b"2.1.1"], [], revision_id=b"2.2.1")
        builder.build_snapshot([b"2.1.2", b"2.2.1"], [], revision_id=b"2.1.3")
        builder.build_snapshot([b"3", b"2.1.3"], [], revision_id=b"4")
        builder.build_snapshot([b"4", b"2.1.2"], [], revision_id=b"5")
        builder.finish_series()
        return builder

    def test_not_an_ancestor(self):
        self.assertRaises(
            errors.CommandError,
            log._generate_all_revisions,
            self.branch,
            "1.1.1",
            "2.1.3",
            "reverse",
            delayed_graph_generation=True,
        )

    def test_wrong_order(self):
        self.assertRaises(
            errors.CommandError,
            log._generate_all_revisions,
            self.branch,
            "5",
            "2.1.3",
            "reverse",
            delayed_graph_generation=True,
        )

    def test_no_start_rev_id_with_end_rev_id_being_a_merge(self):
        log._generate_all_revisions(
            self.branch, None, "2.1.3", "reverse", delayed_graph_generation=True
        )


class TestLogRevSpecsWithPaths(TestLogWithLogCatcher):
    def test_log_revno_n_path_wrong_namespace(self):
        self.make_linear_branch("branch1")
        self.make_linear_branch("branch2")
        # There is no guarantee that a path exist between two arbitrary
        # revisions.
        self.run_bzr("log -r revno:2:branch1..revno:3:branch2", retcode=3)

    def test_log_revno_n_path_correct_order(self):
        self.make_linear_branch("branch2")
        self.assertLogRevnos(["-rrevno:1:branch2..revno:3:branch2"], ["3", "2", "1"])

    def test_log_revno_n_path(self):
        self.make_linear_branch("branch2")
        self.assertLogRevnos(["-rrevno:1:branch2"], ["1"])
        rev_props = self.log_catcher.revisions[0].rev.properties
        self.assertEqual("branch2", rev_props["branch-nick"])


class TestLogErrors(TestLog):
    def test_log_zero_revspec(self):
        self.make_minimal_branch()
        self.run_bzr_error(
            ["brz: ERROR: Logging revision 0 is invalid."], ["log", "-r0"]
        )

    def test_log_zero_begin_revspec(self):
        self.make_linear_branch()
        self.run_bzr_error(
            ["brz: ERROR: Logging revision 0 is invalid."], ["log", "-r0..2"]
        )

    def test_log_zero_end_revspec(self):
        self.make_linear_branch()
        self.run_bzr_error(
            ["brz: ERROR: Logging revision 0 is invalid."], ["log", "-r-2..0"]
        )

    def test_log_nonexistent_revno(self):
        self.make_minimal_branch()
        self.run_bzr_error(
            ["brz: ERROR: Requested revision: '1234' does not exist in branch:"],
            ["log", "-r1234"],
        )

    def test_log_nonexistent_dotted_revno(self):
        self.make_minimal_branch()
        self.run_bzr_error(
            ["brz: ERROR: Requested revision: '123.123' does not exist in branch:"],
            ["log", "-r123.123"],
        )

    def test_log_change_nonexistent_revno(self):
        self.make_minimal_branch()
        self.run_bzr_error(
            ["brz: ERROR: Requested revision: '1234' does not exist in branch:"],
            ["log", "-c1234"],
        )

    def test_log_change_nonexistent_dotted_revno(self):
        self.make_minimal_branch()
        self.run_bzr_error(
            ["brz: ERROR: Requested revision: '123.123' does not exist in branch:"],
            ["log", "-c123.123"],
        )

    def test_log_change_single_revno_only(self):
        self.make_minimal_branch()
        self.run_bzr_error(
            ["brz: ERROR: Option --change does not accept revision ranges"],
            ["log", "--change", "2..3"],
        )

    def test_log_change_incompatible_with_revision(self):
        self.run_bzr_error(
            ["brz: ERROR: --revision and --change are mutually exclusive"],
            ["log", "--change", "2", "--revision", "3"],
        )

    def test_log_nonexistent_file(self):
        self.make_minimal_branch()
        # files that don't exist in either the basis tree or working tree
        # should give an error
        out, err = self.run_bzr("log does-not-exist", retcode=3)
        self.assertContainsRe(
            err, "Path unknown at end or start of revision range: does-not-exist"
        )

    def test_log_reversed_revspecs(self):
        self.make_linear_branch()
        self.run_bzr_error(
            ("brz: ERROR: Start revision must be older than the end revision.\n",),
            ["log", "-r3..1"],
        )

    def test_log_reversed_dotted_revspecs(self):
        self.make_merged_branch()
        self.run_bzr_error(
            ("brz: ERROR: Start revision not found in history of end revision.\n",),
            "log -r 1.1.1..1",
        )

    def test_log_bad_message_re(self):
        """Bad --message argument gives a sensible message.

        See https://bugs.launchpad.net/bzr/+bug/251352
        """
        self.make_minimal_branch()
        out, err = self.run_bzr(["log", "-m", "*"], retcode=3)
        self.assertContainsRe(err, "ERROR.*Invalid pattern.*nothing to repeat")
        self.assertNotContainsRe(err, "Unprintable exception")
        self.assertEqual(out, "")

    def test_log_unsupported_timezone(self):
        self.make_linear_branch()
        self.run_bzr_error(
            [
                'brz: ERROR: Unsupported timezone format "foo", '
                'options are "utc", "original", "local".'
            ],
            ["log", "--timezone", "foo"],
        )

    def test_log_exclude_ancestry_no_range(self):
        self.make_linear_branch()
        self.run_bzr_error(
            ["brz: ERROR: --exclude-common-ancestry requires -r with two revisions"],
            ["log", "--exclude-common-ancestry"],
        )

    def test_log_exclude_ancestry_single_revision(self):
        self.make_merged_branch()
        self.run_bzr_error(
            ["brz: ERROR: --exclude-common-ancestry requires two different revisions"],
            ["log", "--exclude-common-ancestry", "-r1.1.1..1.1.1"],
        )


class TestLogTags(TestLog):
    def test_log_with_tags(self):
        tree = self.make_linear_branch(format="dirstate-tags")
        branch = tree.branch
        branch.tags.set_tag("tag1", branch.get_rev_id(1))
        branch.tags.set_tag("tag1.1", branch.get_rev_id(1))
        branch.tags.set_tag("tag3", branch.last_revision())

        log = self.run_bzr("log -r-1")[0]
        self.assertTrue("tags: tag3" in log)

        log = self.run_bzr("log -r1")[0]
        # I guess that we can't know the order of tags in the output
        # since dicts are unordered, need to check both possibilities
        self.assertContainsRe(log, r"tags: (tag1, tag1\.1|tag1\.1, tag1)")

    def test_merged_log_with_tags(self):
        branch1_tree = self.make_linear_branch("branch1", format="dirstate-tags")
        branch1 = branch1_tree.branch
        branch2_tree = branch1_tree.controldir.sprout("branch2").open_workingtree()
        branch1_tree.commit(message="foobar", allow_pointless=True)
        branch1.tags.set_tag("tag1", branch1.last_revision())
        # tags don't propagate if we don't merge
        self.run_bzr("merge ../branch1", working_dir="branch2")
        branch2_tree.commit(message="merge branch 1")
        log = self.run_bzr("log -n0 -r-1", working_dir="branch2")[0]
        self.assertContainsRe(log, r"    tags: tag1")
        log = self.run_bzr("log -n0 -r3.1.1", working_dir="branch2")[0]
        self.assertContainsRe(log, r"tags: tag1")


class TestLogSignatures(TestLog):
    def test_log_with_signatures(self):
        self.requireFeature(features.gpg)

        self.make_linear_branch(format="dirstate-tags")

        log = self.run_bzr("log --signatures")[0]
        self.assertTrue("signature: no signature" in log)

    def test_log_without_signatures(self):
        self.requireFeature(features.gpg)

        self.make_linear_branch(format="dirstate-tags")

        log = self.run_bzr("log")[0]
        self.assertFalse("signature: no signature" in log)


class TestLogVerbose(TestLog):
    def setUp(self):
        super().setUp()
        self.make_minimal_branch()

    def assertUseShortDeltaFormat(self, cmd):
        log = self.run_bzr(cmd)[0]
        # Check that we use the short status format
        self.assertContainsRe(log, "(?m)^\\s*A  hello.txt$")
        self.assertNotContainsRe(log, "(?m)^\\s*added:$")

    def assertUseLongDeltaFormat(self, cmd):
        log = self.run_bzr(cmd)[0]
        # Check that we use the long status format
        self.assertNotContainsRe(log, "(?m)^\\s*A  hello.txt$")
        self.assertContainsRe(log, "(?m)^\\s*added:$")

    def test_log_short_verbose(self):
        self.assertUseShortDeltaFormat(["log", "--short", "-v"])

    def test_log_s_verbose(self):
        self.assertUseShortDeltaFormat(["log", "-S", "-v"])

    def test_log_short_verbose_verbose(self):
        self.assertUseLongDeltaFormat(["log", "--short", "-vv"])

    def test_log_long_verbose(self):
        # Check that we use the long status format, ignoring the verbosity
        # level
        self.assertUseLongDeltaFormat(["log", "--long", "-v"])

    def test_log_long_verbose_verbose(self):
        # Check that we use the long status format, ignoring the verbosity
        # level
        self.assertUseLongDeltaFormat(["log", "--long", "-vv"])


class TestLogMerges(TestLogWithLogCatcher):
    def setUp(self):
        super().setUp()
        self.make_branches_with_merges()

    def make_branches_with_merges(self):
        level0 = self.make_branch_and_tree("level0")
        self.wt_commit(level0, "in branch level0")
        level1 = level0.controldir.sprout("level1").open_workingtree()
        self.wt_commit(level1, "in branch level1")
        level2 = level1.controldir.sprout("level2").open_workingtree()
        self.wt_commit(level2, "in branch level2")
        level1.merge_from_branch(level2.branch)
        self.wt_commit(level1, "merge branch level2")
        level0.merge_from_branch(level1.branch)
        self.wt_commit(level0, "merge branch level1")

    def test_merges_are_indented_by_level(self):
        self.run_bzr(["log", "-n0"], working_dir="level0")
        [(r.revno, r.merge_depth) for r in self.get_captured_revisions()]
        self.assertEqual(
            [("2", 0), ("1.1.2", 1), ("1.2.1", 2), ("1.1.1", 1), ("1", 0)],
            [(r.revno, r.merge_depth) for r in self.get_captured_revisions()],
        )

    def test_force_merge_revisions_off(self):
        self.assertLogRevnos(["-n1"], ["2", "1"], working_dir="level0")

    def test_force_merge_revisions_on(self):
        self.assertLogRevnos(
            ["-n0"], ["2", "1.1.2", "1.2.1", "1.1.1", "1"], working_dir="level0"
        )

    def test_include_merged(self):
        # Confirm --include-merged gives the same output as -n0
        expected = ["2", "1.1.2", "1.2.1", "1.1.1", "1"]
        self.assertLogRevnos(["--include-merged"], expected, working_dir="level0")
        self.assertLogRevnos(["--include-merged"], expected, working_dir="level0")

    def test_force_merge_revisions_N(self):
        self.assertLogRevnos(
            ["-n2"], ["2", "1.1.2", "1.1.1", "1"], working_dir="level0"
        )

    def test_merges_single_merge_rev(self):
        self.assertLogRevnosAndDepths(
            ["-n0", "-r1.1.2"], [("1.1.2", 0), ("1.2.1", 1)], working_dir="level0"
        )

    def test_merges_partial_range(self):
        self.assertLogRevnosAndDepths(
            ["-n0", "-r1.1.1..1.1.2"],
            [("1.1.2", 0), ("1.2.1", 1), ("1.1.1", 0)],
            working_dir="level0",
        )

    def test_merges_partial_range_ignore_before_lower_bound(self):
        """Dont show revisions before the lower bound's merged revs."""
        self.assertLogRevnosAndDepths(
            ["-n0", "-r1.1.2..2"],
            [("2", 0), ("1.1.2", 1), ("1.2.1", 2)],
            working_dir="level0",
        )

    def test_omit_merges_with_sidelines(self):
        self.assertLogRevnos(
            ["--omit-merges", "-n0"], ["1.2.1", "1.1.1", "1"], working_dir="level0"
        )

    def test_omit_merges_without_sidelines(self):
        self.assertLogRevnos(["--omit-merges", "-n1"], ["1"], working_dir="level0")


class TestLogDiff(TestLogWithLogCatcher):
    # FIXME: We need specific tests for each LogFormatter about how the diffs
    # are displayed: --long indent them by depth, --short use a fixed
    # indent and --line does't display them. -- vila 10019

    def setUp(self):
        super().setUp()
        self.make_branch_with_diffs()

    def make_branch_with_diffs(self):
        level0 = self.make_branch_and_tree("level0")
        self.build_tree(["level0/file1", "level0/file2"])
        level0.add("file1")
        level0.add("file2")
        self.wt_commit(level0, "in branch level0")

        level1 = level0.controldir.sprout("level1").open_workingtree()
        self.build_tree_contents([("level1/file2", b"hello\n")])
        self.wt_commit(level1, "in branch level1")
        level0.merge_from_branch(level1.branch)
        self.wt_commit(level0, "merge branch level1")

    def _diff_file1_revno1(self):
        return b"""=== added file 'file1'
--- file1\t1970-01-01 00:00:00 +0000
+++ file1\t2005-11-22 00:00:00 +0000
@@ -0,0 +1,1 @@
+contents of level0/file1

"""

    def _diff_file2_revno2(self):
        return b"""=== modified file 'file2'
--- file2\t2005-11-22 00:00:00 +0000
+++ file2\t2005-11-22 00:00:01 +0000
@@ -1,1 +1,1 @@
-contents of level0/file2
+hello

"""

    def _diff_file2_revno1_1_1(self):
        return b"""=== modified file 'file2'
--- file2\t2005-11-22 00:00:00 +0000
+++ file2\t2005-11-22 00:00:01 +0000
@@ -1,1 +1,1 @@
-contents of level0/file2
+hello

"""

    def _diff_file2_revno1(self):
        return b"""=== added file 'file2'
--- file2\t1970-01-01 00:00:00 +0000
+++ file2\t2005-11-22 00:00:00 +0000
@@ -0,0 +1,1 @@
+contents of level0/file2

"""

    def assertLogRevnosAndDiff(self, args, expected, working_dir="."):
        self.run_bzr(["log", "-p"] + args, working_dir=working_dir)
        expected_revnos_and_depths = [(revno, depth) for revno, depth, diff in expected]
        # Check the revnos and depths first to make debugging easier
        self.assertEqual(
            expected_revnos_and_depths,
            [(r.revno, r.merge_depth) for r in self.get_captured_revisions()],
        )
        # Now check the diffs, adding the revno  in case of failure
        fmt = "In revno %s\n%s"
        for expected_rev, actual_rev in zip(expected, self.get_captured_revisions()):
            revno, depth, expected_diff = expected_rev
            actual_diff = actual_rev.diff
            self.assertEqualDiff(
                fmt % (revno, expected_diff), fmt % (revno, actual_diff)
            )

    def test_log_diff_with_merges(self):
        self.assertLogRevnosAndDiff(
            ["-n0"],
            [
                ("2", 0, self._diff_file2_revno2()),
                ("1.1.1", 1, self._diff_file2_revno1_1_1()),
                ("1", 0, self._diff_file1_revno1() + self._diff_file2_revno1()),
            ],
            working_dir="level0",
        )

    def test_log_diff_file1(self):
        self.assertLogRevnosAndDiff(
            ["-n0", "file1"],
            [("1", 0, self._diff_file1_revno1())],
            working_dir="level0",
        )

    def test_log_diff_file2(self):
        self.assertLogRevnosAndDiff(
            ["-n1", "file2"],
            [("2", 0, self._diff_file2_revno2()), ("1", 0, self._diff_file2_revno1())],
            working_dir="level0",
        )


class TestLogUnicodeDiff(TestLog):
    def test_log_show_diff_non_ascii(self):
        # Smoke test for bug #328007 UnicodeDecodeError on 'log -p'
        message = "Message with \xb5"
        body = b"Body with \xb5\n"
        wt = self.make_branch_and_tree(".")
        self.build_tree_contents([("foo", body)])
        wt.add("foo")
        wt.commit(message=message)
        # check that command won't fail with unicode error
        # don't care about exact output because we have other tests for this
        out, err = self.run_bzr("log -p --long")
        self.assertNotEqual("", out)
        self.assertEqual("", err)
        out, err = self.run_bzr("log -p --short")
        self.assertNotEqual("", out)
        self.assertEqual("", err)
        out, err = self.run_bzr("log -p --line")
        self.assertNotEqual("", out)
        self.assertEqual("", err)


class TestLogEncodings(tests.TestCaseInTempDir):
    _mu = "\xb5"
    _message = "Message with \xb5"

    # Encodings which can encode mu
    good_encodings = [
        "utf-8",
        "latin-1",
        "iso-8859-1",
        "cp437",  # Common windows encoding
        "cp1251",  # Russian windows encoding
        "cp1258",  # Common windows encoding
    ]
    # Encodings which cannot encode mu
    bad_encodings = [
        "ascii",
        "iso-8859-2",
        "koi8_r",
    ]

    def setUp(self):
        super().setUp()
        self.overrideAttr(osutils, "_cached_user_encoding")

    def create_branch(self):
        brz = self.run_bzr
        brz("init")
        self.build_tree_contents([("a", b"some stuff\n")])
        brz("add a")
        brz(["commit", "-m", self._message])

    def try_encoding(self, encoding, fail=False):
        brz = self.run_bzr
        if fail:
            self.assertRaises(UnicodeEncodeError, self._mu.encode, encoding)
            self._message.encode(encoding, "replace")
        else:
            self._message.encode(encoding)

        old_encoding = osutils._cached_user_encoding
        # This test requires that 'run_bzr' uses the current
        # breezy, because we override user_encoding, and expect
        # it to be used
        try:
            osutils._cached_user_encoding = "ascii"
            # We should be able to handle any encoding
            out, err = brz("log", encoding=encoding)
            if not fail:
                # Make sure we wrote mu as we expected it to exist
                self.assertNotEqual(-1, out.find(self._message))
            else:
                self.assertNotEqual(-1, out.find("Message with ?"))
        finally:
            osutils._cached_user_encoding = old_encoding

    def test_log_handles_encoding(self):
        self.create_branch()

        for encoding in self.good_encodings:
            self.try_encoding(encoding)

    def test_log_handles_bad_encoding(self):
        self.create_branch()

        for encoding in self.bad_encodings:
            self.try_encoding(encoding, fail=True)

    def test_stdout_encoding(self):
        brz = self.run_bzr
        osutils._cached_user_encoding = "cp1251"

        brz("init")
        self.build_tree(["a"])
        brz("add a")
        brz(["commit", "-m", "\u0422\u0435\u0441\u0442"])
        stdout, stderr = self.run_bzr_raw("log", encoding="cp866")

        message = stdout.splitlines()[-1]

        # explanation of the check:
        # u'\u0422\u0435\u0441\u0442' is word 'Test' in russian
        # in cp866  encoding this is string '\x92\xa5\xe1\xe2'
        # in cp1251 encoding this is string '\xd2\xe5\xf1\xf2'
        # This test should check that output of log command
        # encoded to sys.stdout.encoding
        test_in_cp866 = b"\x92\xa5\xe1\xe2"
        test_in_cp1251 = b"\xd2\xe5\xf1\xf2"
        # Make sure the log string is encoded in cp866
        self.assertEqual(test_in_cp866, message[2:])
        # Make sure the cp1251 string is not found anywhere
        self.assertEqual(-1, stdout.find(test_in_cp1251))


class TestLogFile(TestLogWithLogCatcher):
    def test_log_local_branch_file(self):
        """We should be able to log files in local treeless branches."""
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/file"])
        tree.add("file")
        tree.commit("revision 1")
        tree.controldir.destroy_workingtree()
        self.run_bzr("log tree/file")

    def prepare_tree(self, complex=False):
        # The complex configuration includes deletes and renames
        tree = self.make_branch_and_tree("parent")
        self.build_tree(["parent/file1", "parent/file2", "parent/file3"])
        tree.add("file1")
        tree.commit("add file1")
        tree.add("file2")
        tree.commit("add file2")
        tree.add("file3")
        tree.commit("add file3")
        child_tree = tree.controldir.sprout("child").open_workingtree()
        self.build_tree_contents([("child/file2", b"hello")])
        child_tree.commit(message="branch 1")
        tree.merge_from_branch(child_tree.branch)
        tree.commit(message="merge child branch")
        if complex:
            tree.remove("file2")
            tree.commit("remove file2")
            tree.rename_one("file3", "file4")
            tree.commit("file3 is now called file4")
            tree.remove("file1")
            tree.commit("remove file1")
        os.chdir("parent")

    # FIXME: It would be good to parametrize the following tests against all
    # formatters. But the revisions selection is not *currently* part of the
    # LogFormatter contract, so using LogCatcher is sufficient -- vila 100118
    def test_log_file1(self):
        self.prepare_tree()
        self.assertLogRevnos(["-n0", "file1"], ["1"])

    def test_log_file2(self):
        self.prepare_tree()
        # file2 full history
        self.assertLogRevnos(["-n0", "file2"], ["4", "3.1.1", "2"])
        # file2 in a merge revision
        self.assertLogRevnos(["-n0", "-r3.1.1", "file2"], ["3.1.1"])
        # file2 in a mainline revision
        self.assertLogRevnos(["-n0", "-r4", "file2"], ["4", "3.1.1"])
        # file2 since a revision
        self.assertLogRevnos(["-n0", "-r3..", "file2"], ["4", "3.1.1"])
        # file2 up to a revision
        self.assertLogRevnos(["-n0", "-r..3", "file2"], ["2"])

    def test_log_file3(self):
        self.prepare_tree()
        self.assertLogRevnos(["-n0", "file3"], ["3"])

    def test_log_file_historical_missing(self):
        # Check logging a deleted file gives an error if the
        # file isn't found at the end or start of the revision range
        self.prepare_tree(complex=True)
        err_msg = "Path unknown at end or start of revision range: file2"
        err = self.run_bzr("log file2", retcode=3)[1]
        self.assertContainsRe(err, err_msg)

    def test_log_file_historical_end(self):
        # Check logging a deleted file is ok if the file existed
        # at the end the revision range
        self.prepare_tree(complex=True)
        self.assertLogRevnos(["-n0", "-r..4", "file2"], ["4", "3.1.1", "2"])

    def test_log_file_historical_start(self):
        # Check logging a deleted file is ok if the file existed
        # at the start of the revision range
        self.prepare_tree(complex=True)
        self.assertLogRevnos(["file1"], [])

    def test_log_file_renamed(self):
        """File matched against revision range, not current tree."""
        self.prepare_tree(complex=True)

        # Check logging a renamed file gives an error by default
        err_msg = "Path unknown at end or start of revision range: file3"
        err = self.run_bzr("log file3", retcode=3)[1]
        self.assertContainsRe(err, err_msg)

        # Check we can see a renamed file if we give the right end revision
        self.assertLogRevnos(["-r..4", "file3"], ["3"])


class TestLogMultiple(TestLogWithLogCatcher):
    def prepare_tree(self):
        tree = self.make_branch_and_tree("parent")
        self.build_tree(
            [
                "parent/file1",
                "parent/file2",
                "parent/dir1/",
                "parent/dir1/file5",
                "parent/dir1/dir2/",
                "parent/dir1/dir2/file3",
                "parent/file4",
            ]
        )
        tree.add("file1")
        tree.commit("add file1")
        tree.add("file2")
        tree.commit("add file2")
        tree.add(["dir1", "dir1/dir2", "dir1/dir2/file3"])
        tree.commit("add file3")
        tree.add("file4")
        tree.commit("add file4")
        tree.add("dir1/file5")
        tree.commit("add file5")
        child_tree = tree.controldir.sprout("child").open_workingtree()
        self.build_tree_contents([("child/file2", b"hello")])
        child_tree.commit(message="branch 1")
        tree.merge_from_branch(child_tree.branch)
        tree.commit(message="merge child branch")
        os.chdir("parent")

    def test_log_files(self):
        """The log for multiple file should only list revs for those files."""
        self.prepare_tree()
        self.assertLogRevnos(
            ["file1", "file2", "dir1/dir2/file3"], ["6", "5.1.1", "3", "2", "1"]
        )

    def test_log_directory(self):
        """The log for a directory should show all nested files."""
        self.prepare_tree()
        self.assertLogRevnos(["dir1"], ["5", "3"])

    def test_log_nested_directory(self):
        """The log for a directory should show all nested files."""
        self.prepare_tree()
        self.assertLogRevnos(["dir1/dir2"], ["3"])

    def test_log_in_nested_directory(self):
        """The log for a directory should show all nested files."""
        self.prepare_tree()
        os.chdir("dir1")
        self.assertLogRevnos(["."], ["5", "3"])

    def test_log_files_and_directories(self):
        """Logging files and directories together should be fine."""
        self.prepare_tree()
        self.assertLogRevnos(["file4", "dir1/dir2"], ["4", "3"])

    def test_log_files_and_dirs_in_nested_directory(self):
        """The log for a directory should show all nested files."""
        self.prepare_tree()
        os.chdir("dir1")
        self.assertLogRevnos(["dir2", "file5"], ["5", "3"])


class MainlineGhostTests(TestLogWithLogCatcher):
    def setUp(self):
        super().setUp()
        tree = self.make_branch_and_tree("")
        tree.set_parent_ids([b"spooky"], allow_leftmost_as_ghost=True)
        tree.add("")
        tree.commit("msg1", rev_id=b"rev1")
        tree.commit("msg2", rev_id=b"rev2")

    def test_log_range(self):
        self.assertLogRevnos(["-r1..2"], ["2", "1"])

    def test_log_norange(self):
        self.assertLogRevnos([], ["2", "1"])

    def test_log_range_open_begin(self):
        (stdout, stderr) = self.run_bzr(["log", "-r..2"], retcode=3)
        self.assertEqual(["2", "1"], [r.revno for r in self.get_captured_revisions()])
        self.assertEqual("brz: ERROR: Further revision history missing.\n", stderr)

    def test_log_range_open_end(self):
        self.assertLogRevnos(["-r1.."], ["2", "1"])


class TestLogMatch(TestLogWithLogCatcher):
    def prepare_tree(self):
        tree = self.make_branch_and_tree("")
        self.build_tree(["/hello.txt", "/goodbye.txt"])
        tree.add("hello.txt")
        tree.commit(message="message1", committer="committer1", authors=["author1"])
        tree.add("goodbye.txt")
        tree.commit(message="message2", committer="committer2", authors=["author2"])

    def test_message(self):
        self.prepare_tree()
        self.assertLogRevnos(["-m", "message1"], ["1"])
        self.assertLogRevnos(["-m", "message2"], ["2"])
        self.assertLogRevnos(["-m", "message"], ["2", "1"])
        self.assertLogRevnos(["-m", "message1", "-m", "message2"], ["2", "1"])
        self.assertLogRevnos(["--match-message", "message1"], ["1"])
        self.assertLogRevnos(["--match-message", "message2"], ["2"])
        self.assertLogRevnos(["--match-message", "message"], ["2", "1"])
        self.assertLogRevnos(
            ["--match-message", "message1", "--match-message", "message2"], ["2", "1"]
        )
        self.assertLogRevnos(["--message", "message1"], ["1"])
        self.assertLogRevnos(["--message", "message2"], ["2"])
        self.assertLogRevnos(["--message", "message"], ["2", "1"])
        self.assertLogRevnos(
            ["--match-message", "message1", "--message", "message2"], ["2", "1"]
        )
        self.assertLogRevnos(
            ["--message", "message1", "--match-message", "message2"], ["2", "1"]
        )

    def test_committer(self):
        self.prepare_tree()
        self.assertLogRevnos(["-m", "committer1"], ["1"])
        self.assertLogRevnos(["-m", "committer2"], ["2"])
        self.assertLogRevnos(["-m", "committer"], ["2", "1"])
        self.assertLogRevnos(["-m", "committer1", "-m", "committer2"], ["2", "1"])
        self.assertLogRevnos(["--match-committer", "committer1"], ["1"])
        self.assertLogRevnos(["--match-committer", "committer2"], ["2"])
        self.assertLogRevnos(["--match-committer", "committer"], ["2", "1"])
        self.assertLogRevnos(
            ["--match-committer", "committer1", "--match-committer", "committer2"],
            ["2", "1"],
        )

    def test_author(self):
        self.prepare_tree()
        self.assertLogRevnos(["-m", "author1"], ["1"])
        self.assertLogRevnos(["-m", "author2"], ["2"])
        self.assertLogRevnos(["-m", "author"], ["2", "1"])
        self.assertLogRevnos(["-m", "author1", "-m", "author2"], ["2", "1"])
        self.assertLogRevnos(["--match-author", "author1"], ["1"])
        self.assertLogRevnos(["--match-author", "author2"], ["2"])
        self.assertLogRevnos(["--match-author", "author"], ["2", "1"])
        self.assertLogRevnos(
            ["--match-author", "author1", "--match-author", "author2"], ["2", "1"]
        )
